import asyncio
from typing import Any
from rekuest_next.api.schema import PortFragment, PortKind
from rekuest_next.structures.errors import PortShrinkingError, StructureShrinkingError


async def aexpand(port: PortFragment, value: Any, structure_registry=None) -> Any:
    """Expand a value through a port

    Args:
        port (ArgPortFragment): Port to expand to
        value (Any): Value to expand
    Returns:
        Any: Expanded value

    """
    if port.kind == PortKind.DICT:
        return {
            key: await aexpand(port.child, value, structure_registry)
            for key, value in value.items()
        }

    if port.kind == PortKind.LIST:
        return await asyncio.gather(
            *[
                aexpand(port.child, item, structure_registry=structure_registry)
                for item in value
            ]
        )

    if port.kind == PortKind.INT:
        return int(value) if value is not None else int(port.default)

    if port.kind == PortKind.FLOAT:
        return float(value) if value is not None else float(port.default)

    if port.kind == PortKind.STRUCTURE:
        value = value if value is not None else port.default

        if value is None:
            assert port.nullable, "Null value not allowed for non-nullable port"
            return None

        return await structure_registry.get_expander_for_identifier(port.identifier)(
            value
        )

    if port.kind == PortKind.BOOL:
        if value is None:
            value = port.default

        return bool(value) if value is not None else None

    if port.kind == PortKind.STRING:
        if value is None:
            value = port.default

        return str(value) if value is not None else None

    raise NotImplementedError("Should be implemented by subclass")


async def ashrink(port: PortFragment, value: Any, structure_registry=None) -> Any:
    """Expand a value through a port

    Args:
        port (ArgPortFragment): Port to expand to
        value (Any): Value to expand
    Returns:
        Any: Expanded value

    """
    try:
        if value is None:
            if port.nullable:
                return None
            else:
                raise ValueError(
                    "{port} is not nullable (optional) but your provided None"
                )

        if port.kind == PortKind.DICT:
            return {
                key: await ashrink(port.child, value, structure_registry)
                for key, value in value.items()
            }

        if port.kind == PortKind.LIST:
            return await asyncio.gather(
                *[
                    ashrink(port.child, item, structure_registry=structure_registry)
                    for item in value
                ]
            )

        if port.kind == PortKind.INT:
            return int(value) if value is not None else None

        if port.kind == PortKind.STRUCTURE:
            # We always convert structures returns to strings
            try:
                shrinker = structure_registry.get_shrinker_for_identifier(
                    port.identifier
                )
            except KeyError:
                raise StructureShrinkingError(
                    f"Couldn't find shrinker for {port.identifier}"
                ) from None
            try:
                shrink = await shrinker(value)
                return str(shrink)
            except Exception:
                raise StructureShrinkingError(
                    f"Error shrinking {repr(value)} with Structure {port.identifier}"
                ) from None

        if port.kind == PortKind.BOOL:
            return bool(value) if value is not None else None

        if port.kind == PortKind.STRING:
            return str(value) if value is not None else None

        raise NotImplementedError(f"Should be implemented by subclass {port}")

    except Exception as e:
        raise PortShrinkingError(
            f"Couldn't shrink value {value} with port {port}"
        ) from e



from typing import Any, Dict, List, Tuple
import asyncio
from rekuest_next.structures.errors import ExpandingError, ShrinkingError
from rekuest_next.structures.registry import StructureRegistry
import asyncio
from typing import Any, Union
from rekuest_next.api.schema import (
    PortFragment,
    PortKind,
    ChildPortFragment,
    DefinitionInput,
    DefinitionFragment,
)
from rekuest_next.structures.errors import (
    PortShrinkingError,
    StructureShrinkingError,
    StructureExpandingError,
)
from rekuest_next.definition.validate import auto_validate
from .predication import predicate_port
import datetime as dt


async def aexpand_arg(
    port: Union[PortFragment, ChildPortFragment],
    value: Union[str, int, float, dict, list],
    structure_registry,
) -> Any:
    """Expand a value through a port

    Args:
        port (ArgPortFragment): Port to expand to
        value (Any): Value to expand
    Returns:
        Any: Expanded value

    """
    if value is None:
        value = port.default

    if value is None:
        if port.nullable:
            return None
        else:
            raise ExpandingError(
                f"{port.key} is not nullable (optional) but received None"
            )

    if not isinstance(value, (str, int, float, dict, list)):
        raise ExpandingError(
            f"Can't expand {value} of type {type(value)} to {port.kind}. We only accept"
            " strings, ints and floats (json serializable) and null values"
        ) from None

    if port.kind == PortKind.DICT:
        expanding_port = port.children[0]

        if not isinstance(value, dict):
            raise ExpandingError(
                f"Can't expand {value} of type {type(value)} to {port.kind}. We only"
                " accept dicts"
            ) from None

        return {
            key: await aexpand_arg(expanding_port, value, structure_registry)
            for key, value in value.items()
        }

    if port.kind == PortKind.UNION:
        if not isinstance(value, dict):
            raise ExpandingError(
                f"Can't expand {value} of type {type(value)} to {port.kind}. We only"
                " accept dicts in unions"
            )
        assert "use" in value, "No use in vaalue"
        index = value["use"]
        true_value = value["value"]
        return await aexpand_arg(
            port.children[index], true_value, structure_registry=structure_registry
        )

    if port.kind == PortKind.LIST:
        expanding_port = port.children[0]

        if not isinstance(value, list):
            raise ExpandingError(
                f"Can't expand {value} of type {type(value)} to {port.kind}. Only"
                " accept lists"
            ) from None

        return await asyncio.gather(
            *[aexpand_arg(expanding_port, item, structure_registry) for item in value]
        )

    if port.kind == PortKind.MODEL:
        try:
            expanded_args = await asyncio.gather(
                *[
                    aexpand_arg(port, value[port.key], structure_registry)
                    for port in port.children
                ]
            )

            expandend_params = {
                port.key: val for port, val in zip(port.children, expanded_args)
            }

            expander = structure_registry.retrieve_model_expander(port.identifier)
            expanded_values = await expander(expandend_params)
            return expanded_values

        except Exception as e:
            raise ExpandingError(f"Couldn't expand Children {port.children}") from e

    if port.kind == PortKind.INT:
        return int(value)

    if port.kind == PortKind.DATE:
        return dt.datetime.fromisoformat(value.replace("Z", "+00:00"))

    if port.kind == PortKind.FLOAT:
        return float(value)

    if port.kind == PortKind.STRUCTURE:
        try:
            expander = structure_registry.get_expander_for_identifier(port.identifier)
        except KeyError:
            raise StructureExpandingError(
                f"Couldn't find expander for {port.identifier}"
            ) from None

        try:
            expand = await expander(value)
            return expand
        except Exception as e:
            raise StructureExpandingError(
                f"Error expanding {repr(value)} with Structure {port.identifier}"
            ) from e

    if port.kind == PortKind.BOOL:
        return bool(value)

    if port.kind == PortKind.STRING:
        return str(value)

    raise NotImplementedError("Should be implemented by subclass")


async def expand_inputs(
    definition: Union[DefinitionInput, DefinitionFragment],
    args: Dict[str, Union[str, int, float, dict, list]],
    structure_registry: StructureRegistry,
    skip_expanding: bool = False,
):
    """Expand

    Args:
        node (NodeFragment): [description]
        args (List[Any]): [description]
        kwargs (List[Any]): [description]
        registry (Registry): [description]
    """

    expanded_args = []
    node = (
        auto_validate(definition)
        if isinstance(definition, DefinitionInput)
        else definition
    )

    if not skip_expanding:
        try:
            expanded_args = await asyncio.gather(
                *[
                    aexpand_arg(port, args.get(port.key, None), structure_registry)
                    for port in node.args
                ]
            )

            expandend_params = {
                port.key: val for port, val in zip(node.args, expanded_args)
            }

        except Exception as e:
            raise ExpandingError(f"Couldn't expand Arguments {args}") from e
    else:
        expandend_params = {
            port.key: args.get(port.key, None) for port in node.args
        }

    return expandend_params


async def ashrink_return(
    port: Union[PortFragment, ChildPortFragment],
    value: Any,
    structure_registry=None,
) -> Union[str, int, float, dict, list, None]:
    """Expand a value through a port

    Args:
        port (ArgPortFragment): Port to expand to
        value (Any): Value to expand
    Returns:
        Any: Expanded value

    """
    try:
        if value is None:
            if port.nullable:
                return None
            else:
                raise ValueError(
                    f"{port} is not nullable (optional) but your provided None"
                )

        if port.kind == PortKind.UNION:
            for index, x in enumerate(port.variants):
                if predicate_port(x, value, structure_registry):
                    return {
                        "use": index,
                        "value": await ashrink_return(x, value, structure_registry),
                    }

            raise ShrinkingError(
                f"Port is union butn none of the predicated for this port held true {port.variants}"
            )

        if port.kind == PortKind.DICT:
            return {
                key: await ashrink_return(port.child, value, structure_registry)
                for key, value in value.items()
            }

        if port.kind == PortKind.LIST:
            return await asyncio.gather(
                *[
                    ashrink_return(
                        port.child, item, structure_registry=structure_registry
                    )
                    for item in value
                ]
            )

        if port.kind == PortKind.INT:
            return int(value) if value is not None else None

        if port.kind == PortKind.FLOAT:
            return float(value) if value is not None else None

        if port.kind == PortKind.DATE:
            return value.isoformat() if value is not None else None

        if port.kind == PortKind.STRUCTURE:
            # We always convert structures returns to strings
            try:
                shrinker = structure_registry.get_shrinker_for_identifier(
                    port.identifier
                )
            except KeyError:
                raise StructureShrinkingError(
                    f"Couldn't find shrinker for {port.identifier}"
                ) from None
            try:
                shrink = await shrinker(value)
                return str(shrink)
            except Exception as e:
                raise StructureShrinkingError(
                    f"Error shrinking {repr(value)} with Structure {port.identifier}"
                ) from e

        if port.kind == PortKind.BOOL:
            return bool(value) if value is not None else None

        if port.kind == PortKind.STRING:
            return str(value) if value is not None else None

        raise NotImplementedError(f"Should be implemented by subclass {port}")

    except Exception as e:
        raise PortShrinkingError(
            f"Couldn't shrink value {value} with port {port}"
        ) from e


async def shrink_outputs(
    definition: Union[DefinitionInput, DefinitionFragment],
    returns: List[Any],
    structure_registry: StructureRegistry,
    skip_shrinking: bool = False,
) -> Dict[str, Union[str, int, float, dict, list, None]]:
    node = (
        auto_validate(definition)
        if isinstance(definition, DefinitionInput)
        else definition
    )

    if returns is None:
        returns = []
    elif not isinstance(returns, tuple):
        returns = [returns]

    assert len(node.returns) == len(
        returns
    ), (  # We are dealing with a single output, convert it to a proper port like structure
        f"Mismatch in Return Length: expected {len(node.returns)} got {len(returns)}"
    )

    if not skip_shrinking:
        shrinked_returns_future = [
            ashrink_return(port, val, structure_registry)
            for port, val in zip(node.returns, returns)
        ]
        try:
            shrinked_returns = await asyncio.gather(*shrinked_returns_future)
            return {port.key: val for port, val in zip(node.returns, shrinked_returns)}
        except Exception as e:
            raise ShrinkingError(f"Couldn't shrink Returns {returns}") from e
    else:
        return {port.key: val for port, val in zip(node.returns, returns)}



from rekuest_next.postmans.errors import PostmanException


class PostmanTransportException(PostmanException):
    """
    Base class for all exceptions raised by the Agent Transport.
    """


class ReservationListDeniedError(PostmanTransportException):
    """
    Raised when the backend is not able to list the provisions.
    """


class AssignationListDeniedError(PostmanTransportException):
    """
    Raised when the backend is not able to list the assignations.
    """


class AssignDeniedError(PostmanTransportException):
    """
    Raised when the backend is not able to list the provisions.
    """


class UnassignDeniedError(PostmanTransportException):
    """
    Raised when the backend is not able to list the provisions.
    """


class ReserveDeniedError(PostmanTransportException):
    """
    Raised when the backend is not able to list the assignations.
    """


class UnreserveDeniedError(PostmanTransportException):
    """
    Raised when the backend is not able to list the assignations.
    """



from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Optional,
    Type,
    TypeVar,
    Protocol,
    runtime_checkable,
    TYPE_CHECKING,
)
from pydantic import BaseModel
import inspect
from rekuest_next.structures.types import FullFilledStructure
from rekuest_next.api.schema import (
    PortScope,
    AssignWidgetInput,
    ReturnWidgetInput,
    ChoiceInput,
    Identifier,
)
from enum import Enum
from rekuest_next.collection.shelve import get_current_shelve
from .errors import HookError


async def id_shrink(self):
    return self.id


async def shelve_ashrink(cls: Type):
    shelve = get_current_shelve()
    return await shelve.aput(cls)


async def shelve_aexpand(id: str):
    shelve = get_current_shelve()
    return await shelve.aget(id)


async def shelve_acollect(id: str):
    shelve = get_current_shelve()
    return await shelve.adelete(id)


def identity_default_converter(x):
    return x


def cls_to_identifier(cls: Type) -> Identifier:
    return f"{cls.__module__.lower()}.{cls.__name__.lower()}"


def build_instance_predicate(cls: Type):
    return lambda x: isinstance(x, cls)


async def void_acollect(id: str):
    return None


class StandardHookError(HookError):
    pass


class StandardHook(BaseModel):
    cls_to_identifier: Callable[[Type], Identifier] = cls_to_identifier
    """The Standard Hook is a hook that can be registered to the structure registry.

    It will register all local structures in a shelve and will use the shelve to
    expand and shrink the structures. All global structures will net to defined aexpand and
    ashrink using the methods defined in the structure.

    """

    def is_applicable(self, cls: Type) -> bool:
        """Given a class, return True if this hook is applicable to it"""
        return True  # Catch all

    def apply(
        self,
        cls: Type,
        identifier: str = None,
        scope: PortScope = PortScope.LOCAL,
        aexpand: Callable[
            [
                str,
            ],
            Awaitable[Any],
        ] = None,
        ashrink: Callable[
            [
                any,
            ],
            Awaitable[str],
        ] = None,
        acollect: Callable[
            [
                str,
            ],
            Awaitable[Any],
        ] = None,
        predicate: Callable[[Any], bool] = None,
        convert_default: Callable[[Any], str] = None,
        default_widget: Optional[AssignWidgetInput] = None,
        default_returnwidget: Optional[ReturnWidgetInput] = None,
    ) -> FullFilledStructure:
        if identifier is None:
            identifier = self.cls_to_identifier(cls)

        if convert_default is None:
            if hasattr(cls, "convert_default"):
                convert_default = cls.convert_default
            convert_default = identity_default_converter

        if aexpand is None:
            if not hasattr(cls, "aexpand") and scope == PortScope.GLOBAL:
                raise StandardHookError(
                    f"You need to pass 'expand' method or {cls} needs to implement a"
                    " aexpand method if it wants to become a GLOBAL structure"
                )
            aexpand = getattr(cls, "aexpand", shelve_aexpand)

        if ashrink is None:
            if not hasattr(cls, "ashrink") and scope == PortScope.GLOBAL:
                raise StandardHookError(
                    f"You need to pass 'ashrink' method or {cls} needs to implement a"
                    " ashrink method if it wants to become a GLOBAL structure"
                )
            ashrink = getattr(cls, "ashrink", shelve_ashrink)

        if acollect is None:
            if scope == PortScope.GLOBAL:
                acollect = void_acollect
            else:
                acollect = getattr(cls, "acollect", shelve_acollect)

        if predicate is None:
            predicate = build_instance_predicate(cls)

        if identifier is None:
            if not hasattr(cls, "get_identifier"):
                raise StandardHookError(
                    f"You need to pass 'identifier' or  {cls} needs to implement a"
                    " get_identifier method"
                )
            identifier = cls.get_identifier()

        return FullFilledStructure(
            fullfilled_by="StandardHook",
            cls=cls,
            identifier=identifier,
            scope=scope,
            aexpand=aexpand,
            ashrink=ashrink,
            acollect=acollect,
            predicate=predicate,
            convert_default=convert_default,
            default_widget=default_widget,
            default_returnwidget=default_returnwidget,
        )



from rekuest_next.api.schema import (
    AssignWidgetInput,
    ReturnWidgetInput,
    ChoiceInput,
    AssignWidgetKind,
    ReturnWidgetKind,
)
from rekuest_next.scalars import SearchQuery
from typing import List


def SliderWidget(min: int = None, max: int = None, **kwargs) -> AssignWidgetInput:
    """Generate a slider widget.

    Args:
        min (int, optional): The mininum value. Defaults to None.
        max (int, optional): The maximum value. Defaults to None.

    Returns:
        WidgetInput: _description_
    """
    return AssignWidgetInput(kind=AssignWidgetKind.SLIDER, min=min, max=max, **kwargs)


def SearchWidget(query: SearchQuery, ward: str, **kwargs) -> AssignWidgetInput:
    """Generte a search widget.

    A search widget is a widget that allows the user to search for a specifc
    structure utilizing a GraphQL query and running it on a ward (a frontend 
    registered helper that can run the query). The query needs to follow
    the SearchQuery type.

    Args:
        query (SearchQuery): The serach query as a search query object or string
        ward (str): The ward key

    Returns:
        WidgetInput: _description_
    """ """P"""
    return AssignWidgetInput(
        kind=AssignWidgetKind.SEARCH, query=query, ward=ward, **kwargs
    )


def StringWidget(as_paragraph: bool = False, **kwargs) -> AssignWidgetInput:
    """Generate a string widget.

    Args:
        as_paragraph (bool, optional): Should we render the string as a paragraph.Defaults to False.

    Returns:
        WidgetInput: _description_
    """
    return AssignWidgetInput(
        kind=AssignWidgetKind.SEARCH, asParagraph=as_paragraph, **kwargs
    )


def ParagraphWidget(**kwargs) -> AssignWidgetInput:
    """Generate a string widget.

    Args:
        as_paragraph (bool, optional): Should we render the string as a paragraph.Defaults to False.

    Returns:
        WidgetInput: _description_
    """
    return AssignWidgetInput(kind=AssignWidgetKind.SEARCH, asParagraph=True, **kwargs)


def CustomWidget(hook: str, **kwargs) -> AssignWidgetInput:
    """Generate a custom widget.

    A custom widget is a widget that is rendered by a frontend registered hook
    that is passed the input value.

    Args:
        hook (str): The hook key

    Returns:
        WidgetInput: _description_
    """
    return AssignWidgetInput(kind=AssignWidgetKind.CUSTOM, hook=hook, **kwargs)


def CustomReturnWidget(hook: str, **kwargs) -> ReturnWidgetInput:
    """A custom return widget.

    A custom return widget is a widget that is rendered by a frontend registered hook
    that is passed the input value.

    Args:
        hook (str): The hool

    Returns:
        ReturnWidgetInput: _description_
    """ """"""
    return ReturnWidgetInput(kind=ReturnWidgetKind.CUSTOM, hook=hook, **kwargs)


def ChoiceReturnWidget(choices: List[ChoiceInput], **kwargs) -> ReturnWidgetInput:
    """A choice return widget.

    A choice return widget is a widget that renderes a list of choices with the
    value of the choice being highlighted.

    Args:
        choices (List[ChoiceInput]): The choices

    Returns:
        ReturnWidgetInput: _description_
    """
    return ReturnWidgetInput(kind=ReturnWidgetKind.CHOICE, choices=choices, **kwargs)



from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from typing import AsyncIterator, Awaitable, Callable, Optional
from rekuest_next.actors.vars import (
    current_assignment,
    NotWithinAnAssignationError,
)


class SetAssignationLink(ContinuationLink):
    header_name: str = "x-assignation-id"

    async def aconnect(self):
        pass

    async def aexecute(
        self, operation: Operation, **kwargs
    ) -> AsyncIterator[GraphQLResult]:
        try:
            assignment = current_assignment.get()
            operation.context.headers[self.header_name] = assignment.assignation
        except LookupError as e:
            pass

        async for result in self.next.aexecute(operation, **kwargs):
            yield result

    class Config:
        underscore_attrs_are_private = True
        arbitary_types_allowed = True



from abc import abstractmethod
from typing import Any, Awaitable, Callable, List, Optional, Union

from rekuest_next.messages import OutMessage

from koil.composition import KoiledModel
from koil.types import Contextual
from .types import TransportCallbacks


class AgentTransport(KoiledModel):
    """Agent Transport

    A Transport is a means of communicating with an Agent. It is responsible for sending
    and receiving messages from the backend. It needs to implement the following methods:

    list_provision: Getting the list of active provisions from the backend. (depends on the backend)
    list_assignation: Getting the list of active assignations from the backend. (depends on the backend)

    change_assignation: Changing the status of an assignation. (depends on the backend)
    change_provision: Changing the status of an provision. (depends on the backend)

    broadcast: Configuring the callbacks for the transport on new assignation, unassignation provision and unprovison.

    if it is a stateful connection it can also implement the following methods:

    aconnect
    adisconnect

    """

    _callback: Contextual[TransportCallbacks]

    @property
    def connected(self):
        return NotImplementedError("Implement this method")

    @abstractmethod
    async def log_event(self, event: OutMessage):
        raise NotImplementedError("This is an abstract Base Class")

    def set_callback(self, callback: TransportCallbacks):
        self._callback = callback

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True
        copy_on_model_validation = False



from abc import abstractmethod
from typing import Any, Awaitable, Callable, List, Optional, Union

from rekuest_next.messages import OutMessage, ProvisionEvent, AssignationEvent
from rekuest_next.api.schema import ProvisionEventKind, AssignationEventKind, LogLevel

from koil.composition import KoiledModel
from typing import Protocol, runtime_checkable
import logging
import asyncio
from rekuest_next.agents.transport.base import AgentTransport
from rekuest_next.actors.types import Passport, Assignment


logger = logging.getLogger(__name__)


class ProxyAssignTransport(KoiledModel):
    assignment: Assignment
    on_log: Callable[[OutMessage], Awaitable[None]]

    async def log_event(self, *args, **kwargs):
        await self.on_log(
            AssignationEvent(assignation=self.assignment.assignation, **kwargs)
        )  # Forwards assignment up

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True
        copy_on_model_validation = False


class ProxyActorTransport(KoiledModel):
    passport: Passport
    on_log_event: Callable[[OutMessage], Awaitable[None]]

    async def log_event(self, **kwargs):
        await self.on_log_event(
            ProvisionEvent(provision=self.passport.provision, **kwargs)
        )

    def spawn(self, assignment: Assignment) -> ProxyAssignTransport:
        return ProxyAssignTransport(
            assignment=assignment,
            on_log=self.on_log_event,
        )



from pydantic import Field, BaseModel
from rekuest_next.definition.registry import (
    DefinitionRegistry,
    get_default_definition_registry,
)
from rekuest_next.api.schema import TemplateFragment
from rekuest_next.actors.base import Actor, Passport, ActorTransport
from typing import TYPE_CHECKING, Optional

from rekuest_next.agents.errors import ExtensionError


class DefaultExtensionError(ExtensionError):
    pass


if TYPE_CHECKING:
    from rekuest_next.agents.base import BaseAgent


class DefaultExtension(BaseModel):
    definition_registry: DefinitionRegistry = Field(
        default_factory=get_default_definition_registry
    )

    async def should_cleanup_on_init(self) -> bool:
        """Should the extension cleanup its templates?"""
        return True

    async def aspawn_actor_from_template(
        self,
        template: TemplateFragment,
        passport: Passport,
        transport: ActorTransport,
        collector: "Collector",
        agent: "BaseAgent",
    ) -> Optional[Actor]:
        """Spawns an Actor from a Provision. This function closely mimics the
        spawining protocol within an actor. But maps template"""

        try:
            actor_builder = self.definition_registry.get_builder_for_interface(
                template.interface
            )

        except KeyError as e:
            raise ExtensionError(
                f"No Actor Builder found for template {template.interface} and no extensions specified"
            )

        return actor_builder(
            passport=passport,
            transport=transport,
            collector=collector,
            agent=agent,
        )

    async def aretrieve_registry(self):
        return self.definition_registry



from functools import wraps
from typing import List, Dict, Any, Optional, Protocol, runtime_checkable
from pydantic import BaseModel, Field
import asyncio
from .errors import StartupHookError
import inspect


@runtime_checkable
class BackgroundTask(Protocol):
    def __init__(self):
        pass

    async def arun(self, context: Dict[str, Any]): ...


@runtime_checkable
class StartupHook(Protocol):
    def __init__(self):
        pass

    async def arun(self) -> Optional[Dict[str, Any]]:
        """Should return a dictionary of state variables"""
        ...


class HooksRegistry(BaseModel):
    background_worker: Dict[str, BackgroundTask] = Field(default_factory=dict)
    startup_hooks: Dict[str, StartupHook] = Field(default_factory=dict)

    _background_tasks: Dict[str, asyncio.Task] = {}

    def register_background(cls, name: str, task: BackgroundTask):
        assert name not in cls.background_worker, f"Name {name} already registered"
        cls.background_worker[name] = task

    def register_startup(cls, name: str, hook: StartupHook):
        assert name not in cls.startup_hooks, f"Name {name} already registered"
        cls.startup_hooks[name] = hook

    async def arun_startup(self, instance_id: str) -> Dict[str, Any]:
        state = {"instance_id": instance_id}
        for key, hook in self.startup_hooks.items():
            try:
                answer = await hook.arun()
                if answer is not None:
                    state.update(answer)
            except Exception as e:
                raise StartupHookError(f"Startup hook {key} failed") from e
        return state

    async def arun_background(self, context: Dict[str, Any]):
        for name, worker in self.background_worker.items():
            task = asyncio.create_task(worker.arun(context=context))
            task.add_done_callback(lambda x: self._background_tasks.pop(name))
            task.add_done_callback(lambda x: print(f"Worker {name} finished"))
            self._background_tasks[name] = task

    async def astop_background(self):
        for name, task in self._background_tasks.items():
            task.cancel()

        try:
            await asyncio.gather(
                *self._background_tasks.values(), return_exceptions=True
            )
        except asyncio.CancelledError:
            pass

    def reset(self):
        self.background_worker = {}
        self.startup_hooks = {}

    class Config:
        arbitrary_types_allowed = True


default_registry = None


class WrappedStartupHook(StartupHook):
    def __init__(self, func):
        self.func = func

    async def arun(self) -> Optional[Dict[str, Any]]:
        return await self.func()


class WrappedBackgroundTask(BackgroundTask):
    def __init__(self, func):
        self.func = func
        # check if has context argument
        arguments = inspect.signature(func).parameters
        if "context" not in arguments:
            self._do_not_pass_context = True
        else:
            self._do_not_pass_context = False

    async def arun(self, context: Dict[str, Any]):
        if self._do_not_pass_context:
            return await self.func()
        else:
            return await self.func(context)


def get_default_hook_registry() -> HooksRegistry:
    global default_registry
    if default_registry is None:
        default_registry = HooksRegistry()
    return default_registry


def background(
    *func, name: Optional[str] = None, registry: Optional[HooksRegistry] = None
):
    """
    Background tasks are functions that are run in the background
    as asyncio tasks. They are started when the agent starts up
    and stopped automatically when the agent shuts down.

    """

    if len(func) > 1:
        raise ValueError("You can only register one function at a time.")
    if len(func) == 1:
        function = func[0]
        assert asyncio.iscoroutinefunction(
            function
        ), "Startup hooks must be (currently) async"
        registry = registry or get_default_hook_registry()
        name = name or function.__name__
        registry.register_background(name, WrappedBackgroundTask(function))

        return function

    else:

        def real_decorator(function):
            nonlocal registry, name
            assert asyncio.iscoroutinefunction(
                function
            ), "Startup hooks must be (currently) async"

            # Simple bypass for now
            @wraps(function)
            def wrapped_function(*args, **kwargs):
                return function(*args, **kwargs)

            name = name or function.__name__
            registry = registry or get_default_hook_registry()
            registry.register_background(name, WrappedBackgroundTask(function))

            return wrapped_function

        return real_decorator


def startup(
    *func, name: Optional[str] = None, registry: Optional[HooksRegistry] = None
):
    """
    This is a decorator that registers a function as a startup hook.
    Startup hooks are called when the agent starts up and AFTER the
    definitions have been registered with the agent.

    Then, the startup hook is called and the state variables are
    returned as a dictionary. These state variables are then passed
    accessible in every actors' context.
    """
    if len(func) > 1:
        raise ValueError("You can only register one function at a time.")
    if len(func) == 1:
        function = func[0]
        assert asyncio.iscoroutinefunction(
            function
        ), "Startup hooks must be (currently) async"
        registry = registry or get_default_hook_registry()
        name = name or function.__name__

        registry.register_startup(name, WrappedStartupHook(function))

        return function

    else:

        def real_decorator(function):
            nonlocal registry, name
            assert asyncio.iscoroutinefunction(
                function
            ), "Startup hooks must be (currently) async"

            # Simple bypass for now
            @wraps(function)
            def wrapped_function(*args, **kwargs):
                return function(*args, **kwargs)

            registry = registry or get_default_hook_registry()
            name = name or function.__name__
            registry.register_startup(name, WrappedStartupHook(function))

            return wrapped_function

        return real_decorator



from typing import Awaitable, Callable, Dict, Any, List, Optional
import websockets
from rekuest_next.agents.transport.base import AgentTransport
import asyncio
import json
from rekuest_next.agents.transport.errors import (
    AgentTransportException,
    AssignationListDeniedError,
    ProvisionListDeniedError,
)
from rekuest_next.messages import (
    Assign,
    OutMessage,
    Provide,
    Unprovide,
    Cancel,
    Init,
    Interrupt,
    MessageType,
    Message,
)
import logging
from websockets.exceptions import (
    ConnectionClosedError,
    InvalidStatusCode,
    InvalidHandshake,
)
from pydantic import Field
from rekuest_next.api.schema import AssignationEventKind, LogLevel, ProvisionEventKind
import ssl
import certifi
from koil.types import ContextBool, Contextual
from .errors import (
    CorrectableConnectionFail,
    DefiniteConnectionFail,
    AgentWasKicked,
    AgentIsAlreadyBusy,
    AgentWasBlocked,
)

logger = logging.getLogger(__name__)


async def token_loader():
    raise NotImplementedError(
        "Websocket transport does need a defined token_loader on Connection"
    )


KICK_CODE = 3001
BUSY_CODE = 3002
BLOCKED_CODE = 3003
BOUNCED_CODE = 3004

agent_error_codes = {
    KICK_CODE: AgentWasKicked,
    BUSY_CODE: AgentIsAlreadyBusy,
    BLOCKED_CODE: AgentWasBlocked,
}

agent_error_message = {
    KICK_CODE: "Agent was kicked by the server",
    BUSY_CODE: "Agent can't connect on this instance_id as another instance is already connected. Please kick the other instance first or use another instance_id",
    BLOCKED_CODE: "Agent was blocked by the server",
}


class WebsocketAgentTransport(AgentTransport):
    endpoint_url: str
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where())
    )
    token_loader: Callable[[], Awaitable[str]] = Field(exclude=True)
    max_retries = 5
    time_between_retries = 3
    allow_reconnect = True
    auto_connect = True

    _futures: Contextual[Dict[str, asyncio.Future]] = None
    _connected: ContextBool = False
    _healthy: ContextBool = False
    _send_queue: Contextual[asyncio.Queue] = None
    _connection_task: Contextual[asyncio.Task] = None
    _connected_future: Contextual[asyncio.Future]

    async def __aenter__(self):
        assert (
            self._callback is not None
        ), "Callback not set. Use set callback first to set it"
        self._futures = {}

        self._send_queue = asyncio.Queue()

    async def aconnect(self, instance_id: str):
        self._connected_future = asyncio.Future()
        self._connection_task = asyncio.create_task(self.websocket_loop(instance_id))
        self._connected = await self._connected_future

    async def on_definite_error(self, e: Exception):
        if not self._connected_future.done():
            self._connected_future.set_exception(e)
        else:
            return await self._callback.on_definite_error(e)

    async def abroadcast(self, *args, **kwargs):
        await self._callback.abroadcast(*args, **kwargs)

    async def on_agent_error(self, e: Exception):
        if not self._connected_future.done():
            self._connected_future.set_exception(e)
        else:
            return await self._callback.on_agent_error(e)

    async def on_correctable_error(self, e: Exception):
        return await self._callback.on_correctable_error(e)

    async def websocket_loop(self, instance_id: str, retry=0, reload_token=False):
        send_task = None
        receive_task = None
        try:
            try:
                token = await self.token_loader(force_refresh=reload_token)

                async with websockets.connect(
                    f"{self.endpoint_url}",
                    ssl=(
                        self.ssl_context
                        if self.endpoint_url.startswith("wss")
                        else None
                    ),
                ) as client:
                    retry = 0
                    logger.info("Agent on Websockets connected")

                    await client.send(
                        json.dumps(
                            {
                                "type": "INITIAL",
                                "token": token,
                                "instance_id": instance_id,
                            }
                        )
                    )

                    send_task = asyncio.create_task(self.sending(client))
                    receive_task = asyncio.create_task(self.receiving(client))

                    self._healthy = True
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self._healthy = False

                    for task in pending:
                        task.cancel()

                    for task in done:
                        raise task.exception()

            except InvalidHandshake as e:
                logger.warning(
                    (
                        "Websocket to"
                        f" {self.endpoint_url}?token={token}&instance_id={instance_id} was"
                        " denied. Trying to reload token"
                    ),
                    exc_info=True,
                )
                reload_token = True
                raise CorrectableConnectionFail from e

            except ConnectionClosedError as e:
                logger.warning("Websocket was closed", exc_info=True)
                if e.code in agent_error_codes:
                    await self.on_agent_error(
                        agent_error_codes[e.code](agent_error_message[e.code])
                    )
                    raise AgentTransportException("Agent Error") from e

                if e.code == BOUNCED_CODE:
                    raise CorrectableConnectionFail(
                        "Was bounced. Debug call to reconnect"
                    ) from e

                else:
                    raise CorrectableConnectionFail(
                        "Connection failed unexpectably. Reconnectable."
                    ) from e

            except Exception as e:
                logger.error("Websocket excepted closed definetely", exc_info=True)
                await self.on_definite_error(DefiniteConnectionFail(str(e)))
                logger.critical("Unhandled exception... ", exc_info=True)
                raise DefiniteConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(f"Trying to Recover from Exception {e}")

            should_retry = await self._callback.on_correctable_error(e)

            if retry > self.max_retries or not self.allow_reconnect or not should_retry:
                logger.error("Max retries reached. Giving up")
                raise DefiniteConnectionFail("Exceeded Number of Retries")

            logger.info(
                f"Waiting for some time before retrying: {self.time_between_retries}"
            )
            await asyncio.sleep(self.time_between_retries)
            logger.info("Retrying to connect")
            await self.websocket_loop(
                instance_id, retry=retry + 1, reload_token=reload_token
            )

        except asyncio.CancelledError as e:
            logger.info("Websocket got cancelled. Trying to shutdown graceully")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

            await asyncio.gather(send_task, receive_task, return_exceptions=True)
            raise e

    async def sending(self, client):
        try:
            while True:
                message = await self._send_queue.get()
                logger.info(f">>>> {message}")
                await client.send(message)
                self._send_queue.task_done()
        except asyncio.CancelledError:
            logger.info("Sending Task sucessfully Cancelled")

    async def receiving(self, client):
        try:
            async for message in client:
                logger.info(f"Receiving message {message}")
                await self.receive(message)

        except asyncio.CancelledError:
            logger.info("Receiving Task sucessfully Cancelled")

    async def receive(self, message):
        json_dict = json.loads(message)
        logger.debug(f"<<<< {message}")
        if "type" in json_dict:
            type = json_dict["type"]
            id = json_dict["id"]

            if type == MessageType.HEARTBEAT:
                await self._send_queue.put(json.dumps({"type": "HEARTBEAT"}))

            # State Layer
            if type == MessageType.INIT:
                initial_message = Init(**json_dict)

                if not self._connected_future.done():
                    self._connected_future.set_result(True)

                for i in initial_message.provisions:
                    await self.abroadcast(i)

                for i in initial_message.inquiries:
                    await self.abroadcast(i)

            if type == MessageType.ASSIGN:
                await self.abroadcast(Assign(**json_dict))

            if type == MessageType.CANCEL:
                await self.abroadcast(Cancel(**json_dict))

            if type == MessageType.UNPROVIDE:
                await self.abroadcast(Unprovide(**json_dict))

            if type == MessageType.PROVIDE:
                await self.abroadcast(Provide(**json_dict))

            if type == MessageType.INTERRUPT:
                await self.abroadcast(Interrupt(**json_dict))

        else:
            logger.error(f"Unexpected messsage: {json_dict}")

    async def awaitaction(self, action: Message):
        assert self._connected, "Should be connected"
        if action.id in self._futures:
            raise ValueError("Action already has a future")

        future = asyncio.Future()
        self._futures[action.id] = future
        await self._send_queue.put(action.json())
        return await future

    async def delayaction(self, action: Message):
        assert self._connected, "Should be connected"
        await self._send_queue.put(action.json())

    async def log_event(self, event: OutMessage):
        await self.delayaction(event)

    async def adisconnect(self):
        self._connection_task.cancel()
        self._connected = False

        try:
            await self._connection_task
        except:
            pass

        self._connection_task = None

    async def __aexit__(self, *args, **kwargs):
        if self._connection_task:
            await self.adisconnect()

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True
        copy_on_model_validation = "none"



from pydantic import Field
from rath import rath
import contextvars
from rath.links.auth import AuthTokenLink

from rath.links.compose import TypedComposedLink
from rath.links.dictinglink import DictingLink
from rath.links.shrink import ShrinkingLink
from rath.links.split import SplitLink
from rath.links.retry import RetryLink

current_rekuest_next_rath = contextvars.ContextVar(
    "current_rekuest_next_rath", default=None
)


class RekuestNextLinkComposition(TypedComposedLink):
    shrink: ShrinkingLink = Field(default_factory=ShrinkingLink)
    dicting: DictingLink = Field(default_factory=DictingLink)
    auth: AuthTokenLink
    retry: RetryLink = Field(default_factory=RetryLink)
    split: SplitLink

    def _repr_html_inline_(self):
        return f"<table><tr><td>auth</td><td>{self.auth.maximum_refresh_attempts}</td></tr></table>"


class RekuestNextRath(rath.Rath):
    link: RekuestNextLinkComposition = Field(default_factory=RekuestNextLinkComposition)

    async def __aenter__(self):
        await super().__aenter__()
        current_rekuest_next_rath.set(self)
        return self

    def _repr_html_inline_(self):
        return f"<table><tr><td>link</td><td>{self.link._repr_html_inline_()}</td></tr></table>"

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super().__aexit__(exc_type, exc_val, exc_tb)
        current_rekuest_next_rath.set(None)



from typing import runtime_checkable, Protocol, Optional
from rekuest_next.api.schema import TemplateFragment
from rekuest_next.actors.base import Actor, Passport, ActorTransport
from typing import TYPE_CHECKING
from rekuest_next.definition.registry import DefinitionRegistry
from rekuest_next.collection.collector import Collector

if TYPE_CHECKING:
    from rekuest_next.agents.base import BaseAgent


@runtime_checkable
class AgentExtension(Protocol):
    async def should_cleanup_on_init(self) -> bool:
        """Should the extension cleanup its templates?"""
        ...

    async def get_name(self) -> str:
        """This should return the name of the extension"""
        raise NotImplementedError("Implement this method")

    async def aspawn_actor_from_template(
        self,
        template: TemplateFragment,
        passport: Passport,
        transport: ActorTransport,
        agent: "BaseAgent",
        collector: "Collector",
    ) -> Optional[Actor]:
        """This should create an actor from a template and return it.

        The actor should not be started!

        TODO: This should be asserted

        """
        ...

    async def aretrieve_registry(
        self,
    ) -> DefinitionRegistry:
        """This should register the definitions for the agent.

        This is called when the agent is started, for each extensions. Extensions
        should register their definitions here and merge them with the agent's
        definition registry.


        Basic usage:

        ```python

        definition, actorBuilder = reactify(...)

        definition_registry.register_at_interface(
            "deploy_graph",
            definition=definition,
            structure_registry=self.structure_registry,
            actorBuilder=actorBuilder,
        )

        ```

        Merge Usage:

        ```python

        self.definition_registry = DefinitionRegistry()

        agent_definition_registry.merge_with(self.definition_registry)



        """
        ...



import hashlib
import json

from rekuest_next.api.schema import DefinitionFragment, DefinitionInput


def auto_validate(defintion: DefinitionInput) -> DefinitionFragment:
    """Validates a definition against its own schema

    This should always be the first step in the validation process
    but does not guarantee that the definition is valid on the connected
    arkitekt service. That means that the definition might be valid
    within this client (e.g. you can access and assign to actors in this
    context, but the arkitekt service might not be able to adress your actor
    or assign to it.)

    """

    hm = defintion.dict(by_alias=True)
    del hm["interfaces"]

    return DefinitionFragment(**hm)


def hash_definition(definition: DefinitionInput):
    hashable_definition = {
        key: value
        for key, value in dict(definition.dict()).items()
        if key not in ["meta", "interface"]
    }
    return hashlib.sha256(
        json.dumps(hashable_definition, sort_keys=True).encode()
    ).hexdigest()



from typing import Callable
from pydantic import BaseModel, root_validator, validator
import uuid
import random


class PortTrait(BaseModel):
    """
    Class for validating port input
    on the client side

    """

    @validator("default", check_fields=False)
    def validator(cls, v):
        # Check if the default value is JSON serializable
        if v is None:
            return v

        if not isinstance(v, (str, int, float, dict, list, bool)):
            raise ValueError(
                "Default value must be JSON serializable, got: " + str(v)
            ) from None

        return v

    @root_validator(pre=True)
    def validate_portkind_nested(cls, values):
        from rekuest_next.api.schema import PortKind

        kind = values.get("kind")

        if kind is None:
            raise ValueError("kind is required")

        if kind == PortKind.STRUCTURE:
            if values.get("identifier") is None:
                raise ValueError(
                    "When specifying a structure you need to provide an arkitekt"
                    " identifier got:" + str(values)
                )

        if kind == PortKind.LIST:
            if values.get("children") is None:
                raise ValueError(
                    "When specifying a list you need to provide a wrapped 'chidlren' port"
                )

        return values

    def mock(
        self,
        structure_generator: Callable = uuid.uuid4,
        int_generator: Callable = lambda: random.randint(0, 100),
        float_generator: Callable = lambda: random.random(),
        string_generator: Callable = lambda: str("sss"),
    ):
        """
        Mocks some serialized data for this port
        """
        from rekuest_next.api.schema import PortKind

        kind = self.kind

        if kind == PortKind.STRUCTURE:
            return str(structure_generator())

        if kind == PortKind.LIST:
            return [self.child.mock()]

        if kind == PortKind.DICT:
            return {"hello": self.child.mock(), "world": self.child.mock()}

        if kind == PortKind.STRING:
            return string_generator()

        if kind == PortKind.INT:
            return int_generator()

        if kind == PortKind.BOOL:
            return float_generator()

        return None


class WidgetInputTrait(BaseModel):
    """
    Class for validating widget input
    on the client side

    """

    @root_validator(pre=True)
    def validate_widgetkind_nested(cls, values):
        from rekuest_next.api.schema import WidgetKind

        kind = values.get("kind")

        if kind is None:
            raise ValueError("kind is required")

        if kind == WidgetKind.SearchWidget:
            if values.get("query") is None:
                raise ValueError(
                    "When specifying a SearchWidget you need to provide an query"
                    " parameter"
                )

        if kind == WidgetKind.SliderWidget:
            if values.get("min") is None or values.get("max") is None:
                raise ValueError(
                    "When specifying a Slider you need to provide an 'max and 'min'"
                    f" parameter {values}"
                )

            if values.get("min") > values.get("max"):
                raise ValueError(
                    "When specifying a Slider you need to provide an 'max' greater than"
                    " 'min'"
                )

        return values


class ReturnWidgetInputTrait(BaseModel):
    """
    Class for validating widget input
    on the client side

    """

    @root_validator(pre=True)
    def validate_widgetkind_nested(cls, values):
        from rekuest_next.api.schema import ReturnWidgetKind

        kind = values.get("kind")

        if kind is None:
            raise ValueError("kind is required")

        if kind == ReturnWidgetKind.CUSTOM:
            if values.get("hook") is None:
                raise ValueError(
                    "When specifying a CustomReturnWidget you need to provide a 'hook'"
                    " parameter, corresponding to the desired reigstered hook"
                )

        return values


class AnnotationInputTrait(BaseModel):
    """
    Abstract class for serialization of data.

    """

    @root_validator(pre=True)
    def validate_annotationkind_nested(cls, values):
        from rekuest_next.api.schema import AnnotationKind

        kind = values.get("kind")

        if kind is None:
            raise ValueError("kind is required")

        if kind == AnnotationKind.ValueRange:
            if values.get("min") is None and values.get("max") is None:
                raise ValueError("min or max is required when using Value Range")

            if values.get("min") is not None and values.get("max") is not None:
                if values.get("min") > values.get("max"):
                    raise ValueError(
                        "When using a ValueRange min must be less than max"
                    )

        if kind == AnnotationKind.IsPredicate:
            if values.get("predicate") is None:
                raise ValueError("predicate is required when using IsPredicate")

        if kind == AnnotationKind.CustomAnnotation:
            if values.get("hook") is None:
                raise ValueError("hook is required when using CustomAnnotation")

        if kind == AnnotationKind.AttributePredicate:
            if values.get("attribute") is None:
                raise ValueError("atrribute is required when using AttributePredicate")
            if values.get("annotations") is None:
                raise ValueError(
                    "annotations on the predicate is required when using"
                    " AttributePredicate"
                )

        return values



from typing import Any, Awaitable, Callable, Dict, Optional
from pydantic import Field
from rekuest_next.postmans.transport.websocket import WebsocketPostmanTransport
from fakts.fakt.base import Fakt
from herre import Herre
from fakts import Fakts


class WebsocketPostmanTransportConfig(Fakt):
    endpoint_url: str
    instance_id: str = "default"


class ArkitektWebsocketPostmanTransport(WebsocketPostmanTransport):
    endpoint_url: Optional[str]
    instance_id: Optional[str]
    token_loader: Optional[Callable[[], Awaitable[str]]] = Field(exclude=True)
    fakts: Fakts
    herre: Herre
    fakts_group: str
    _old_fakt: Dict[str, Any] = {}

    def configure(self, fakt: WebsocketPostmanTransportConfig) -> None:
        self.endpoint_url = fakt.endpoint_url
        self.instance_id = fakt.instance_id
        self.token_loader = self.token_loader or self.herre.aget_token

    async def aconnect(self):
        if self.fakts.has_changed(self._old_fakt, self.fakts_group):
            self._old_fakt = await self.fakts.aget(self.fakts_group)
            self.configure(WebsocketPostmanTransportConfig(**self._old_fakt))

        return await super().aconnect()



try:
    from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
    from rath.links.split import SplitLink
    from rath.contrib.fakts.links.graphql_ws import FaktsGraphQLWSLink
    from rath.contrib.herre.links.auth import HerreAuthLink
    from rekuest_next.rath import RekuestNextLinkComposition, RekuestNextRath
    from rekuest_next.rekuest import RekuestNext
    from graphql import OperationType
    from rekuest_next.contrib.arkitekt.websocket_agent_transport import (
        ArkitektWebsocketAgentTransport,
    )
    from reaktion_next.extension import ReaktionExtension

    from rekuest_next.agents.base import BaseAgent
    from fakts import Fakts
    from herre import Herre
    from rekuest_next.postmans.graphql import GraphQLPostman
    from arkitekt_next.service_registry import (
        get_default_service_builder_registry,
        Params,
    )

    from .structures.default import get_default_structure_registry
    from .api.schema import (
            AssignationEventFragment,
            aget_event,
            NodeFragment,
            afind,
            Search_nodesQuery,
    )
    from rekuest_next.structures.hooks.standard import id_shrink
    from rekuest_next.widgets import SearchWidget
    from arkitekt_next.model import Requirement

    class ArkitektNextRekuestNext(RekuestNext):
        rath: RekuestNextRath
        agent: BaseAgent

    def builder(fakts: Fakts, herre: Herre, params: Params) -> ArkitektNextRekuestNext:
        instance_id = params.get("instance_id", "default")

        rath = RekuestNextRath(
            link=RekuestNextLinkComposition(
                auth=HerreAuthLink(herre=herre),
                split=SplitLink(
                    left=FaktsAIOHttpLink(fakts_group="rekuest", fakts=fakts),
                    right=FaktsGraphQLWSLink(fakts_group="rekuest", fakts=fakts),
                    split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                ),
            )
        )

        agent = BaseAgent(
            transport=ArkitektWebsocketAgentTransport(
                fakts_group="rekuest.agent", fakts=fakts, herre=herre
            ),
            instance_id=instance_id,
            rath=rath,
        )

        try:
            from reaktion_next.extension import ReaktionExtension

            agent.extensions["reaktion"] = ReaktionExtension()
        except ImportError as e:
            raise e

        return ArkitektNextRekuestNext(
            rath=rath,
            agent=agent,
            postman=GraphQLPostman(
                rath=rath,
                instance_id=instance_id,
            ),
        )

    service_builder_registry = get_default_service_builder_registry()
    service_builder_registry.register(
        "rekuest",
        builder,
        Requirement(
            service="live.arkitekt.rekuest",
            description="An instance of ArkitektNext Rekuest to assign to nodes",
        ),
    )


    structur_reg = get_default_structure_registry()


    structur_reg.register_as_structure(
        AssignationEventFragment,
        identifier="@rekuest/assignationevent",
        aexpand=aget_event,
        ashrink=id_shrink,
    )

    structur_reg.register_as_structure(
        NodeFragment,
        identifier="@rekuest/node",
        aexpand=afind,
        ashrink=id_shrink,
        default_widget=SearchWidget(query=Search_nodesQuery.Meta.document, ward="rekuest"),
    )

    imported = True


except ImportError as e:
    imported = False



from typing import Any, Dict, List, Optional, Union
from rekuest_next.api.schema import (
    AssignationFragment,
    AssignationStatus,
    ReservationFragment,
    ReservationStatus,
)

from rekuest_next.messages import Assignation, Reservation
from rekuest_next.postmans.base import BasePostman
import asyncio
from pydantic import Field
import logging
from .transport.base import PostmanTransport

logger = logging.getLogger(__name__)


class StatefulPostman(BasePostman):
    transport: PostmanTransport
    assignations: Dict[str, Assignation] = Field(default_factory=dict)
    reservations: Dict[str, Reservation] = Field(default_factory=dict)

    _res_update_queues: Dict[str, asyncio.Queue] = {}
    _ass_update_queues: Dict[str, asyncio.Queue] = {}

    async def aconnect(self):
        await super().aconnect()

        data = await self.transport.alist_reservations()
        self.reservations = {res.reservation: res for res in data}

        data = await self.transport.alist_assignations()
        self.assignations = {ass.assignation: ass for ass in data}

    async def areserve(
        self,
        node: str,
        params: dict = None,
        provision: str = None,
        reference: str = None,
    ) -> asyncio.Queue:
        return await self.transport.areserve(
            node, params, provision=provision, reference=reference
        )

    async def aunreserve(self, reservation_id: str) -> ReservationFragment:
        unreservation = await self.transport.aunreserve(reservation_id)
        self.reservations[
            unreservation.reservation
        ].status = ReservationStatus.CANCELING
        return self.reservations[unreservation.reservation]

    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        parent: Optional[str] = None,
        persist=True,
        log=False,
    ) -> asyncio.Queue:
        return await self.transport.aassign(reservation, args, persist, log)

    async def aunassign(
        self,
        assignation: str,
    ) -> AssignationFragment:
        unassignation = await self.transport.aunassign(assignation)
        self.assignations[
            unassignation.assignation
        ].status = AssignationStatus.CANCELING
        return unassignation

    def register_reservation_queue(
        self, node: str, reference: str, queue: asyncio.Queue
    ):
        self._res_update_queues[node + reference] = queue

    def register_assignation_queue(self, ass_id: str, queue: asyncio.Queue):
        self._ass_update_queues[ass_id] = queue

    async def abroadcast(self, message: Union[Assignation, Reservation]):
        if isinstance(message, Assignation):
            if message.assignation in self._ass_update_queues:
                self.assignations[message.assignation].update(message)
                await self._ass_update_queues[message.assignation].put(
                    self.assignations[message.assignation]
                )
            else:
                logger.warning(
                    "Received Assignation Update without having knowingly queued it."
                    " Most likely because client crashed before receiving updates.  We"
                    " will omit!"
                )
        elif isinstance(message, Reservation):
            if message.reservation in self._res_update_queues:
                self.reservations[message.reservation].update(message)
                await self._res_update_queues[message.reservation].put(
                    self.reservations[message.reservation]
                )
            else:
                logger.warning(
                    "Received Reservation Update without having knowingly queued it."
                    " Most likely because client crashed before receiving updates. We"
                    " will omit!"
                )

        else:
            raise Exception("Unknown message type")

    def unregister_reservation_queue(self, node: str, reference: str):
        del self._res_update_queues[node + reference]

    def unregister_assignation_queue(self, ass_id: str):
        del self._ass_update_queues[ass_id]

    async def __aenter__(self):
        await self.transport.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.transport.__aexit__(exc_type, exc_val, exc_tb)

    class Config:
        arbitrary_types_allowed = True



try:
    from typing import Annotated, get_type_hints, Any

    annot_type = type(Annotated[int, "spam"])

    def is_annotated(obj: Any) -> bool:
        """Checks if a hint is an Annotated type

        Args:
            hint (Any): The typehint to check
            annot_type (_type_, optional): _description_. Defaults to annot_type.

        Returns:
            bool: _description_
        """
        return (type(obj) is annot_type) and hasattr(obj, "__metadata__")

except ImportError:
    Annotated = None
    from typing import get_type_hints as _get_type_hints, Any

    def get_type_hints(obj: Any, include_extras=False, **kwargs):
        return _get_type_hints(obj, **kwargs)

    def is_annotated(obj: Any) -> bool:
        """Checks if a hint is an Annotated type

        Args:
            hint (Any): The typehint to check
            annot_type (_type_, optional): _description_. Defaults to annot_type.

        Returns:
            bool: _description_
        """
        return False



import asyncio
import uuid
from typing import Any, AsyncGenerator, List, Optional, Protocol, Union

from koil import unkoil, unkoil_gen
from koil.composition.base import KoiledModel

from rekuest_next.actors.vars import (
    get_current_assignation_helper,
    NotWithinAnAssignationError,
)
from rekuest_next.api.schema import (
    AssignationEventKind,
    AssignInput,
    HookInput,
    NodeFragment,
    ReservationFragment,
    ReserveInput,
    TemplateFragment,
    UnreserveInput,
    afind,
    areserve,
    aunreserve,
)
from rekuest_next.postmans.base import BasePostman
from rekuest_next.postmans.vars import get_current_postman
from rekuest_next.structures.registry import get_current_structure_registry
from rekuest_next.structures.serialization.postman import aexpand_returns, ashrink_args


def useUser() -> str:
    """Use the current User

    Returns:
        User: The current User
    """
    helper = get_current_assignation_helper()
    return helper.assignation.user


class IDBearer(Protocol):
    id: str


def none_or_id(value: Optional[Union[IDBearer, str]]) -> Optional[str]:
    if value:
        if isinstance(value, str):
            return value
        else:
            if hasattr(value, "id"):
                return value.id
            else:
                raise ValueError("The object has no id")
    return None


async def acall(
    *args,
    node: Optional[Union[NodeFragment, str]] = None,
    template: Optional[Union[TemplateFragment, str]] = None,
    reservation: Optional[Union[ReservationFragment, str]] = None,
    reference: Optional[str] = None,
    hooks: Optional[List[HookInput]] = None,
    cached: bool = False,
    parent: bool = None,
    log: bool = False,
    **kwargs,
) -> tuple[Any]:
    if not isinstance(node, NodeFragment):
        node = await afind(id=node)

    postman: BasePostman = get_current_postman()
    structure_registry = get_current_structure_registry()

    shrinked_args = await ashrink_args(
        node, args, kwargs, structure_registry=structure_registry
    )

    instance_id = postman.instance_id

    try:
        parent = parent or get_current_assignation_helper().assignation
    except NotWithinAnAssignationError as e:
        print("Not in assignation")
        parent = None

    reference = reference or str(uuid.uuid4())

    value = []
    async for i in postman.aassign(
        AssignInput(
            instanceId=instance_id,
            node=none_or_id(node),
            template=none_or_id(template),
            reservation=none_or_id(reservation),  # type: ignore
            args=shrinked_args,
            reference=reference,
            hooks=hooks or [],
            cached=cached,
            parent=parent,
            log=log,
            isHook=False,
        )
    ):
        print(i)
        if i.kind == AssignationEventKind.YIELD:
            value = i.returns

        if i.kind == AssignationEventKind.DONE:
            return await aexpand_returns(
                node, value, structure_registry=structure_registry
            )


async def aiterate(
    *args,
    node: Optional[NodeFragment] = None,
    template: Optional[TemplateFragment] = None,
    reservation: Optional[ReservationFragment] = None,
    reference: Optional[str] = None,
    hooks: Optional[List[HookInput]] = None,
    cached: bool = False,
    parent: bool = None,
    log: bool = False,
    **kwargs,
) -> AsyncGenerator[tuple[Any], None]:
    if not isinstance(node, NodeFragment):
        node = await afind(id=node)

    postman: BasePostman = get_current_postman()
    structure_registry = get_current_structure_registry()

    shrinked_args = await ashrink_args(
        node, args, kwargs, structure_registry=structure_registry
    )

    instance_id = postman.instance_id

    try:
        parent = parent or get_current_assignation_helper().assignation
    except:
        print("Not in assignation")
        parent = None

    reference = reference or str(uuid.uuid4())

    async for i in postman.aassign(
        AssignInput(
            instanceId=instance_id,
            node=node.id,
            args=shrinked_args,
            reference=reference,
            hooks=hooks or [],
            cached=cached,
            parent=parent,
            log=log,
            isHook=False,
        )
    ):
        if i.kind == AssignationEventKind.YIELD:
            yield await aexpand_returns(
                node, i.returns, structure_registry=structure_registry
            )

        if i.kind == AssignationEventKind.DONE:
            break

        if i.kind == AssignationEventKind.CRITICAL:
            raise Exception(i.returns)


async def aiterate_raw(
    node: Optional[NodeFragment] = None,
    template: Optional[TemplateFragment] = None,
    reservation: Optional[ReservationFragment] = None,
    reference: Optional[str] = None,
    hooks: Optional[List[HookInput]] = None,
    cached: bool = False,
    parent: bool = None,
    log: bool = False,
    **kwargs,
) -> AsyncGenerator[tuple[Any], None]:
    postman: BasePostman = get_current_postman()

    instance_id = postman.instance_id

    try:
        parent = parent or get_current_assignation_helper().assignation
    except:
        print("Not in assignation")
        parent = None

    reference = reference or str(uuid.uuid4())

    x = AssignInput(
        instanceId=instance_id,
        node=none_or_id(node),
        template=none_or_id(template),
        reservation=none_or_id(reservation),  # type: ignore
        args=kwargs,
        reference=reference,
        hooks=hooks or [],
        cached=cached,
        parent=parent,
        log=log,
        isHook=False,
    )

    print("assogm omüit", x)

    async for i in postman.aassign(x):
        if i.kind == AssignationEventKind.YIELD:
            yield i.returns

        if i.kind == AssignationEventKind.DONE:
            break

        if i.kind == AssignationEventKind.CRITICAL:
            raise Exception(i.message)


async def acall_raw(
    node: Optional[NodeFragment] = None,
    template: Optional[TemplateFragment] = None,
    reservation: Optional[ReservationFragment] = None,
    reference: Optional[str] = None,
    hooks: Optional[List[HookInput]] = None,
    cached: bool = False,
    parent: bool = None,
    log: bool = False,
    **kwargs,
) -> AsyncGenerator[tuple[Any], None]:
    postman: BasePostman = get_current_postman()

    instance_id = postman.instance_id

    try:
        parent = parent or get_current_assignation_helper().assignation
    except:
        print("Not in assignation")
        parent = None

    reference = reference or str(uuid.uuid4())

    x = AssignInput(
        instanceId=instance_id,
        node=none_or_id(node),
        template=none_or_id(template),
        reservation=none_or_id(reservation),  # type: ignore
        args=kwargs,
        reference=reference,
        hooks=hooks or [],
        cached=cached,
        parent=parent,
        log=log,
        isHook=False,
    )

    print("assogm omüit", x)

    returns = None

    async for i in postman.aassign(x):
        if i.kind == AssignationEventKind.YIELD:
            returns = i.returns

        if i.kind == AssignationEventKind.DONE:
            return returns

        if i.kind == AssignationEventKind.CRITICAL:
            raise Exception(i.message)


def call(
    *args,
    node: Optional[NodeFragment] = None,
    template: Optional[TemplateFragment] = None,
    reservation: Optional[ReservationFragment] = None,
    reference: Optional[str] = None,
    hooks: Optional[List[HookInput]] = None,
    cached: bool = False,
    parent: bool = None,
    log: bool = False,
    **kwargs,
) -> tuple[Any]:
    return unkoil(
        acall,
        node,
        reference=reference,
        hooks=hooks,
        cached=cached,
        parent=parent,
        log=log,
        **kwargs,
    )


def iterate(
    *args,
    node: Optional[NodeFragment] = None,
    template: Optional[TemplateFragment] = None,
    reservation: Optional[ReservationFragment] = None,
    reference: Optional[str] = None,
    hooks: Optional[List[HookInput]] = None,
    cached: bool = False,
    parent: bool = None,
    log: bool = False,
    **kwargs,
) -> tuple[Any]:
    return unkoil_gen(
        aiterate,
        node,
        reference=reference,
        hooks=hooks,
        cached=cached,
        parent=parent,
        log=log,
        **kwargs,
    )


class ReservationContext(KoiledModel):
    node: NodeFragment
    constants: Optional[dict[str, Any]] = None
    reference: Optional[str] = None
    hooks: Optional[List[HookInput]] = None
    cached: bool = False
    parent: Optional[str] = None
    log: bool = False
    assignation_id: Optional[str] = None

    _reservation: Optional[ReservationFragment] = None

    async def __aenter__(self) -> "ReservationContext":
        _postman = get_current_postman()

        self._reservation = await areserve(
            ReserveInput(
                node=self.node.id,
                assignationId=self.assignation_id,
                instanceId=_postman.instance_id,
            )
        )

        return self

    async def acall(self, *args, **kwargs):
        assert self._reservation, "Not in reservation context"
        return await acall(
            node=self.node, reservation=self._reservation, *args, **kwargs
        )

    async def acall_raw(self, *args, **kwargs):
        assert self._reservation, "Not in reservation context"
        return await acall_raw(
            node=self.node, reservation=self._reservation, *args, **kwargs
        )

    def call(self, *args, **kwargs):
        return unkoil(self.acall, *args, **kwargs)

    async def aiterate(self, *args, **kwargs):
        assert self._reservation, "Not in reservation context"
        async for i in aiterate(
            node=self.node, reservation=self._reservation, *args, **kwargs
        ):
            yield i

    async def aiterate_raw(self, *args, **kwargs):
        assert self._reservation, "Not in reservation context"
        async for i in aiterate_raw(
            node=self.node, reservation=self._reservation, *args, **kwargs
        ):
            yield i

    def iterate(self, *args, **kwargs):
        return unkoil_gen(self.aiterate, *args, **kwargs)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        from rekuest_next.api.schema import ReserveInput, UnreserveInput, aunreserve

        if self._reservation:
            await aunreserve(UnreserveInput(reservation=self._reservation.id))

        return self

    def __enter__(self) -> "ReservationContext":
        return super().__enter__()

    class Config:
        underscore_attrs_are_private = True


def reserved(
    node: NodeFragment,
    reference: Optional[str] = None,
    hooks: Optional[List[HookInput]] = None,
    cached: bool = False,
    parent: bool = None,
    log: bool = False,
    constants: Optional[dict[str, Any]] = None,
    assignation_id: Optional[str] = None,
) -> ReservationContext:
    try:
        assignation_id = assignation_id or get_current_assignation_helper().assignation
    except NotWithinAnAssignationError:
        assignation_id = None

    return ReservationContext(
        node=node,
        reference=reference,
        hooks=hooks,
        cached=cached,
        parent=parent,
        log=log,
        constants=constants,
        assignation_id=assignation_id,
    )



import types
import typing
from typing import get_origin


UNION_TYPE_PREDICATE = getattr(types, "UnionType", False)


def cls_is_union(cls):
    return get_origin(cls) is UNION_TYPE_PREDICATE or get_origin(cls) is typing.Union


def cls_is_list(cls):
    return get_origin(cls) is typing.List



from typing import Any, Tuple
from qtpy import QtCore
from rekuest_next.agents.transport.base import AgentTransport
from rekuest_next.messages import Provision
from koil.qt import QtCoro
from rekuest_next.actors.functional import FunctionalFuncActor
from qtpy import QtWidgets
from rekuest_next.definition.registry import ActorBuilder
from rekuest_next.definition.define import prepare_definition, DefinitionInput
from rekuest_next.actors.types import ActorBuilder, Passport
from rekuest_next.collection.collector import ActorCollector
from rekuest_next.actors.transport.types import ActorTransport


class QtInLoopBuilder(QtCore.QObject):
    """A function that takes a provision and an actor transport and returns an actor.

    The actor produces by this builder will be running in the same thread as the
    koil instance (aka, the thread that called the builder).

    Args:
        QtCore (_type_): _description_
    """

    def __init__(
        self,
        assign=None,
        *args,
        parent=None,
        structure_registry=None,
        definition=None,
        **actor_kwargs
    ) -> None:
        super().__init__(*args, parent=parent)
        self.coro = QtCoro(
            lambda *args, **kwargs: assign(*args, **kwargs), autoresolve=True
        )
        self.provisions = {}
        self.structure_registry = structure_registry
        self.actor_kwargs = actor_kwargs
        self.definition = definition

    async def on_assign(self, *args, **kwargs) -> None:
        return await self.coro.acall(*args, **kwargs)

    async def on_provide(self, provision: Provision) -> Any:
        return None

    async def on_unprovide(self) -> Any:
        return None

    def build(self, *args, **kwargs) -> Any:
        try:
            ac = FunctionalFuncActor(
                *args,
                **kwargs,
                structure_registry=self.structure_registry,
                assign=self.on_assign,
                on_provide=self.on_provide,
                on_unprovide=self.on_unprovide,
                definition=self.definition,
            )
            return ac
        except Exception as e:
            raise e


class QtFutureBuilder(QtCore.QObject):
    """A function that takes a provision and an actor transport and returns an actor.

    The actor produces by this builder will be running in the same thread as the
    koil instance (aka, the thread that called the builder).

    Args:
        QtCore (_type_): _description_
    """

    def __init__(
        self,
        assign=None,
        *args,
        parent=None,
        structure_registry=None,
        definition=None,
        **actor_kwargs
    ) -> None:
        super().__init__(*args, parent=parent)
        self.coro = QtCoro(
            lambda *args, **kwargs: assign(*args, **kwargs), autoresolve=False
        )
        self.provisions = {}
        self.structure_registry = structure_registry
        self.actor_kwargs = actor_kwargs
        self.definition = definition

    async def on_assign(self, *args, **kwargs) -> None:
        x = await self.coro.acall(*args, **kwargs)
        return x

    async def on_provide(self, provision: Provision) -> Any:
        return None

    async def on_unprovide(self) -> Any:
        return None

    def build(self, *args, **kwargs) -> Any:
        try:
            ac = FunctionalFuncActor(
                *args,
                **kwargs,
                structure_registry=self.structure_registry,
                assign=self.on_assign,
                on_provide=self.on_provide,
                on_unprovide=self.on_unprovide,
                definition=self.definition,
            )
            return ac
        except Exception as e:
            raise e


def qtinloopactifier(
    function, structure_registry, parent: QtWidgets.QWidget = None, **kwargs
) -> Tuple[DefinitionInput, ActorBuilder]:
    """Qt Actifier

    The qt actifier wraps a function and returns a builder that will create an actor
    that runs in the same thread as the Qt instance, enabling the use of Qt widgets
    and signals.
    """

    definition = prepare_definition(function, structure_registry, **kwargs)

    in_loop_instance = QtInLoopBuilder(
        parent=parent,
        assign=function,
        structure_registry=structure_registry,
        definition=definition,
    )

    def builder(*args, **kwargs) -> Any:
        return in_loop_instance.build(
            *args, **kwargs
        )  # build an actor for this inloop instance

    return definition, builder


def qtwithfutureactifier(
    function, structure_registry, parent: QtWidgets.QWidget = None, **kwargs
) -> ActorBuilder:
    """Qt Actifier

    The qt actifier wraps a function and returns a build that calls the function with
    its first parameter being a future that can be resolved within the qt loop
    """

    definition = prepare_definition(function, structure_registry, omitfirst=1, **kwargs)

    in_loop_instance = QtFutureBuilder(
        parent=parent,
        assign=function,
        structure_registry=structure_registry,
        definition=definition,
    )

    def builder(*args, **kwargs) -> Any:
        return in_loop_instance.build(
            *args, **kwargs
        )  # build an actor for this inloop instance

    return definition, builder



from rekuest_next.actors.types import Actifier
from rekuest_next.definition.validate import hash_definition
from rekuest_next.structures.registry import (
    StructureRegistry,
    get_current_structure_registry,
)
from rekuest_next.structures.default import (
    get_default_structure_registry,
)
from rekuest_next.definition.registry import (
    DefinitionRegistry,
    get_default_definition_registry,
    get_current_definition_registry,
)
from rekuest_next.api.schema import (
    AssignWidgetInput,
    DependencyInput,
    PortGroupInput,
    PortScope,
    ReturnWidgetInput,
    EffectInput,
    TemplateInput,
    ValidatorFunction,
    ValidatorInput,
    CreateTemplateInput,
)
from rekuest_next.collection.shelve import get_current_shelve
from typing import Dict, List, Callable, Optional, Tuple, Awaitable, Any
import inflection
from rekuest_next.definition.define import prepare_definition
from rekuest_next.actors.actify import reactify
from functools import wraps, partial


def register_func(
    function_or_actor,
    structure_registry: StructureRegistry,
    definition_registry: DefinitionRegistry,
    interface: str = None,
    actifier: Actifier = reactify,
    dependencies: List[DependencyInput] = None,
    port_groups: Optional[List[PortGroupInput]] = None,
    groups: Optional[Dict[str, List[str]]] = None,
    validators: Optional[Dict[str, List[ValidatorInput]]] = None,
    collections: List[str] = None,
    is_test_for: Optional[List[str]] = None,
    logo: Optional[str] = None,
    widgets: Dict[str, AssignWidgetInput] = None,
    effects: Dict[str, List[EffectInput]] = None,
    interfaces: List[str] = [],
    on_provide=None,
    on_unprovide=None,
    dynamic: bool = False,
    in_process: bool = False,
    **actifier_params,
):
    """Register a function or actor with the definition registry

    Register a function or actor with the definition registry. This will
    create a definition for the function or actor and register it with the
    definition registry.

    If first parameter is a function, it will be wrapped in an actorBuilder
    through the actifier. If the first parameter is an actor, it will be
    used as the actorBuilder (needs to have the dunder __definition__) to be
    detected as such.

    Args:
        function_or_actor (Union[Actor, Callable]): _description_
        actifier (Actifier, optional): _description_. Defaults to None.
        interface (str, optional): _description_. Defaults to None.
        widgets (Dict[str, WidgetInput], optional): _description_. Defaults to {}.
        interfaces (List[str], optional): _description_. Defaults to [].
        on_provide (_type_, optional): _description_. Defaults to None.
        on_unprovide (_type_, optional): _description_. Defaults to None.
        structure_registry (StructureRegistry, optional): _description_. Defaults to None.
    """

    interface = interface or inflection.underscore(
        function_or_actor.__name__
    )  # convert this to camelcase

    
    definition, actor_builder = actifier(
        function_or_actor,
        structure_registry,
        on_provide=on_provide,
        on_unprovide=on_unprovide,
        widgets=widgets,
        is_test_for=is_test_for,
        collections=collections,
        groups=groups,
        logo=logo,
        port_groups=port_groups,
        effects=effects,
        validators=validators,
        interfaces=interfaces,
        in_process=in_process,
        **actifier_params,
    )

    definition_registry.register_at_interface(
        interface,
        TemplateInput(
            interface=interface,
            definition=definition,
            dependencies=dependencies or [],
            logo=logo,
            dynamic=dynamic,
        ),
        structure_registry,
        actor_builder,
    )

    return definition, actor_builder


def register(
    *func,
    actifier: Actifier = reactify,
    interface: str = None,
    widgets: Dict[str, AssignWidgetInput] = None,
    dependencies: List[DependencyInput] = None,
    interfaces: List[str] = [],
    collections: List[str] = None,
    port_groups: Optional[List[PortGroupInput]] = None,
    groups: Optional[Dict[str, List[str]]] = None,
    effects: Dict[str, List[EffectInput]] = None,
    is_test_for: Optional[List[str]] = None,
    logo: Optional[str] = None,
    on_provide=None,
    on_unprovide=None,
    validators: Optional[Dict[str, List[ValidatorInput]]] = None,
    structure_registry: StructureRegistry = None,
    definition_registry: DefinitionRegistry = None,
    in_process: bool = False,
    dynamic: bool = False,
    **actifier_params,
):
    """Register a function or actor to the default definition registry.

    You can use this decorator to register a function or actor to the default
    definition registry. There is also a function version of this decorator,
    which is more convenient to use.

    Example:
        >>> @register
        >>> def hello_world(string: str):

        >>> @register(interface="hello_world")
        >>> def hello_world(string: str):

    Args:
        function_or_actor (Union[Callable, Actor]): The function or Actor
        builder (ActorBuilder, optional): An actor builder (see ActorBuilder). Defaults to None.
        package (str, optional): The package you want to register this function in. Defaults to standard app package    .
        interface (str, optional): The name of the function. Defaults to the functions name.
        widgets (Dict[str, WidgetInput], optional): A dictionary of parameter key and a widget. Defaults to the default widgets as registered in the structure registry .
        interfaces (List[str], optional): Interfaces that this node adheres to. Defaults to [].
        on_provide (Callable[[ProvisionFragment], Awaitable[dict]], optional): Function that shall be called on provide (in the async eventloop). Defaults to None.
        on_unprovide (Callable[[], Awaitable[dict]], optional): Function that shall be called on unprovide (in the async eventloop). Defaults to None.
        structure_registry (StructureRegistry, optional): The structure registry to use for this Actor (used to shrink and expand inputs). Defaults to None.
    """
    definition_registry = definition_registry or get_default_definition_registry()
    structure_registry = structure_registry or get_default_structure_registry()

    if len(func) > 1:
        raise ValueError("You can only register one function or actor at a time.")
    if len(func) == 1:
        function_or_actor = func[0]

        @wraps(function_or_actor)
        def wrapped_function(*args, **kwargs):
            return function_or_actor(*args, **kwargs)

        definition, actor_builder = register_func(
            function_or_actor,
            structure_registry=structure_registry,
            definition_registry=definition_registry,
            dependencies=dependencies,
            validators=validators,
            actifier=actifier,
            interface=interface,
            is_test_for=is_test_for,
            widgets=widgets,
            logo=logo,
            effects=effects,
            collections=collections,
            interfaces=interfaces,
            on_provide=on_provide,
            on_unprovide=on_unprovide,
            port_groups=port_groups,
            groups=groups,
            in_process=in_process,
            dynamic=dynamic,
            **actifier_params,
        )

        wrapped_function.__definition__ = definition
        wrapped_function.__definition_hash__ = hash_definition(definition)
        wrapped_function.__actor_builder__ = actor_builder

        return wrapped_function

    else:

        def real_decorator(function_or_actor):
            # Simple bypass for now
            @wraps(function_or_actor)
            def wrapped_function(*args, **kwargs):
                return function_or_actor(*args, **kwargs)

            definition, actor_builder = register_func(
                function_or_actor,
                structure_registry=structure_registry,
                definition_registry=definition_registry,
                actifier=actifier,
                interface=interface,
                validators=validators,
                dependencies=dependencies,
                is_test_for=is_test_for,
                widgets=widgets,
                effects=effects,
                collections=collections,
                interfaces=interfaces,
                on_provide=on_provide,
                logo=logo,
                on_unprovide=on_unprovide,
                port_groups=port_groups,
                groups=groups,
                dynamic=dynamic,
                in_process=in_process,
                **actifier_params,
            )

            wrapped_function.__definition__ = definition
            wrapped_function.__definition_hash__ = hash_definition(definition)
            wrapped_function.__actor_builder__ = actor_builder

            return wrapped_function

        return real_decorator


def register_structure(
    *cls,
    identifier: str = None,
    scope: PortScope = PortScope.LOCAL,
    aexpand: Callable[
        [
            str,
        ],
        Awaitable[Any],
    ] = None,
    ashrink: Callable[
        [
            any,
        ],
        Awaitable[str],
    ] = None,
    acollect: Callable[
        [
            str,
        ],
        Awaitable[Any],
    ] = None,
    convert_default: Callable[[Any], str] = None,
    default_widget: AssignWidgetInput = None,
    default_returnwidget: ReturnWidgetInput = None,
    registry: StructureRegistry = None,
    **kwargs,
):
    """Register a structure to the default structure registry.

    Args:
        cls (Structure): The structure class
        name (str, optional): The name of the structure. Defaults to the class name.
    """
    if len(cls) > 1:
        raise ValueError("You can only register one function or actor at a time.")
    if len(cls) == 1:
        function_or_actor = cls[0]

        sregistry = registry or get_default_structure_registry()

        sregistry.register_as_structure(
            function_or_actor,
            identifier=identifier,
            scope=scope,
            ashrink=ashrink,
            aexpand=aexpand,
            acollect=acollect,
            convert_default=convert_default,
            default_widget=default_widget,
            default_returnwidget=default_returnwidget,
            **kwargs,
        )

        return cls

    else:

        def real_decorator(cls):
            # Simple bypass for now

            sregistry = registry or get_default_structure_registry()

            sregistry.register_as_structure(
                cls,
                identifier=identifier,
                scope=scope,
                ashrink=ashrink,
                aexpand=aexpand,
                acollect=acollect,
                convert_default=convert_default,
                default_widget=default_widget,
                default_returnwidget=default_returnwidget,
                **kwargs,
            )

            return cls

        return real_decorator



from typing import Any, AsyncGenerator, Dict, List, Union
import uuid
from rekuest_next.api.schema import (
    AssignationEventFragment,
    AssignationFragment,
    ReservationFragment,
    aassign,
    areserve,
    awatch_assignations,
    awatch_reservations,
    acancel,
    aunreserve,
    AssignInput,
    CancelInput,
    AssignationChangeEventFragment,
    InterruptInput,
    BindsInput,
)
from rekuest_next.postmans.base import BasePostman
import asyncio
from pydantic import Field
import logging
from .errors import PostmanException
from .vars import current_postman
from rekuest_next.rath import RekuestNextRath

logger = logging.getLogger(__name__)


class GraphQLPostman(BasePostman):
    rath: RekuestNextRath
    instance_id: str
    assignations: Dict[str, AssignationFragment] = Field(default_factory=dict)

    _ass_update_queues: Dict[str, asyncio.Queue] = {}

    _ass_update_queue: asyncio.Queue = None

    _watch_assraces_task: asyncio.Task = None
    _watch_assignations_task: asyncio.Task = None

    _watching: bool = None
    _lock: asyncio.Lock = None

    async def aconnect(self):
        await super().aconnect()
        data = {}  # await self.transport.alist_assignations()
        self.assignations = {ass.assignation: ass for ass in data}

    async def aunreserve(self, reservation_id: str):
        async with self._lock:
            if not self._watching:
                await self.start_watching()

        try:
            unreservation = await aunreserve(reservation_id)
            self.reservations[unreservation.id] = unreservation
        except Exception as e:
            raise PostmanException("Cannot Unreserve") from e

    async def aassign(
        self, assign: AssignInput
    ) -> AsyncGenerator[AssignationEventFragment, None]:
        async with self._lock:
            if not self._watching:
                await self.start_watching()

        print(assign)

        try:
            assignation = await aassign(assign)
            print("Assigned on postman")
        except Exception as e:
            raise PostmanException("Cannot Assign") from e

        self._ass_update_queues[assign.reference] = asyncio.Queue()
        queue = self._ass_update_queues[assign.reference]

        try:
            while True:
                signal = await queue.get()
                yield signal
                queue.task_done()

        except asyncio.CancelledError as e:
            unassignation = await acancel(CancelInput(assignation=assignation.id))
            del self._ass_update_queues[assign.reference]
            raise e

    def register_assignation_queue(self, ass_id: str, queue: asyncio.Queue):
        self._ass_update_queues[ass_id] = queue

    def unregister_assignation_queue(self, ass_id: str):
        del self._ass_update_queues[ass_id]

    async def watch_assignations(self):
        print("Start to watch")
        try:
            async for assignation in awatch_assignations(
                self.instance_id, rath=self.rath
            ):
                if assignation.event:
                    reference = assignation.event.reference
                    await self._ass_update_queues[reference].put(assignation.event)
                if assignation.create:
                    print("Created assignation")
                    if assignation.create.reference not in self._ass_update_queues:
                        print("NOOOOOT IN QUEUES, RACE CONDITION")

        except Exception as e:
            print(e)

    async def watch_assraces(self):
        try:
            while True:
                ass: AssignationEventFragment = await self._ass_update_queue.get()
                self._ass_update_queue.task_done()
                logger.info(f"Postman received Assignation {ass}")

                unique_identifier = ass.reference

                await self._ass_update_queues[unique_identifier].put(ass)

        except Exception:
            logger.error("Error in watch_resraces", exc_info=True)

    async def start_watching(self):
        logger.info("Starting watching")
        self._ass_update_queue = asyncio.Queue()
        self._watch_assignations_task = asyncio.create_task(self.watch_assignations())
        self._watch_assignations_task.add_done_callback(self.log_assignation_fail)
        self._watch_assraces_task = asyncio.create_task(self.watch_assraces())
        self._watching = True

    def log_assignation_fail(self, future):
        print(future)
        return

    async def stop_watching(self):
        self._watch_assignations_task.cancel()
        self._watch_assraces_task.cancel()

        try:
            await asyncio.gather(
                self._watch_assignations_task,
                self._watch_assraces_task,
                return_exceptions=True,
            )
        except asyncio.CancelledError:
            pass

        self._watching = False

    async def __aenter__(self):
        self._lock = asyncio.Lock()
        current_postman.set(self)
        return await super().__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._watching:
            await self.stop_watching()
        current_postman.set(None)
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    class Config:
        underscore_attrs_are_private = True



from dataclasses import dataclass
from fieldz import fields, Field
from typing import List, Dict
from .types import FullFilledArg, FullFilledModel
import inflection


def model(cls):
    """Mark a class as a model. This is used to
    identify the model in the rekuest_next system."""

    try:
        fields(cls)
    except TypeError:
        raise TypeError(
            "Models must be serializable by fieldz in order to be used in rekuest_next."
        )

    setattr(cls, "__rekuest_model__", inflection.underscore(cls.__name__))

    return dataclass(cls)


def is_model(cls):
    """Check if a class is a model."""

    return getattr(cls, "__rekuest_model__", False)


def retrieve_args_for_model(cls) -> List[FullFilledArg]:
    children_clses = fields(cls)

    args = []
    for field in children_clses:
        print(field.default)
        print(field)
        args.append(
            FullFilledArg(
                cls=field.type,
                default=field.default if field.default != Field.MISSING else None,
                key=field.name,
                description=field.description,
            )
        )
    print(args)
    return args


def retrieve_fullfiled_model(cls) -> FullFilledModel:
    return FullFilledModel(
        identifier=cls.__rekuest_model__,
        description=cls.__doc__,
        args=retrieve_args_for_model(cls),
    )



import asyncio
import logging
import uuid
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from pydantic import Field

from koil import unkoil
from koil.composition import KoiledModel
from rekuest_next.actors.base import Actor
from rekuest_next.actors.transport.local_transport import (
    ProxyActorTransport,
)
from rekuest_next.actors.transport.types import ActorTransport
from rekuest_next.actors.types import ActorBuilder, Assignment, Passport, Unassignment
from rekuest_next.agents.errors import AgentException, ProvisionException
from rekuest_next.agents.extension import AgentExtension
from rekuest_next.agents.transport.base import AgentTransport, Contextual
from rekuest_next.api.schema import (
    AssignationEventKind,
    ProvisionEventKind,
    TemplateFragment,
    acreate_template,
    aget_provision,
    CreateTemplateInput,
    aensure_agent,
    SetExtensionTemplatesInput,
    aset_extension_templates,
    acreate_hardware_record,
)
from rekuest_next.collection.collector import Collector
from rekuest_next.definition.registry import (
    DefinitionRegistry,
    get_current_definition_registry,
    get_default_definition_registry,
)
from rekuest_next.agents.hooks import HooksRegistry, get_default_hook_registry
from rekuest_next.definition.validate import auto_validate
from rekuest_next.messages import (
    Assign,
    InMessage,
    OutMessage,
    Cancel,
    Interrupt,
    Message,
    Provide,
    Unprovide,
    AssignationEvent,
    AssignInquiry,
    ProvisionEvent,
)
from rekuest_next.rath import RekuestNextRath
from rekuest_next.agents.extensions.default import DefaultExtension
from .transport.errors import CorrectableConnectionFail, DefiniteConnectionFail

logger = logging.getLogger(__name__)


def build_base_extensions():
    return {
        "default": DefaultExtension(),
    }


class BaseAgent(KoiledModel):
    """Agent

    Agents are the governing entities for every app. They are responsible for
    managing the lifecycle of the direct actors that are spawned from them through arkitekt.

    Agents are nothing else than actors in the classic distributed actor model, but they are
    always provided when the app starts and they do not provide functionality themselves but rather
    manage the lifecycle of the actors that are spawned from them.

    The actors that are spawned from them are called guardian actors and they are the ones that+
    provide the functionality of the app. These actors can then in turn spawn other actors that
    are not guardian actors. These actors are called non-guardian actors and their lifecycle is
    managed by the guardian actors that spawned them. This allows for a hierarchical structure
    of actors that can be spawned from the agents.


    """

    instance_id: str = "main"
    rath: RekuestNextRath
    transport: AgentTransport
    extensions: Dict[str, AgentExtension] = Field(default_factory=build_base_extensions)
    collector: Collector = Field(default_factory=Collector)
    managed_actors: Dict[str, Actor] = Field(default_factory=dict)
    interface_template_map: Dict[str, TemplateFragment] = Field(default_factory=dict)
    template_interface_map: Dict[str, str] = Field(default_factory=dict)
    provision_passport_map: Dict[str, Passport] = Field(default_factory=dict)
    managed_assignments: Dict[str, Assignment] = Field(default_factory=dict)
    hook_registry: HooksRegistry = Field(default_factory=get_default_hook_registry)
    _provisionTaskMap: Dict[str, asyncio.Task] = Field(default_factory=dict)
    _inqueue: Contextual[asyncio.Queue] = None
    _errorfuture: Contextual[asyncio.Future] = None
    _context: Dict[str, Any] = None

    started = False
    running = False

    async def abroadcast(self, message: InMessage):
        await self._inqueue.put(message)

    async def on_agent_error(self, exception) -> None:
        if self._errorfuture is None or self._errorfuture.done():
            return
        self._errorfuture.set_exception(exception)
        ...

    async def on_definite_error(self, error: DefiniteConnectionFail) -> None:
        if self._errorfuture is None or self._errorfuture.done():
            return
        self._errorfuture.set_exception(error)
        ...

    async def on_correctable_error(self, error: CorrectableConnectionFail) -> bool:
        # Always correctable
        return True
        ...

    async def process(self, message: InMessage):
        logger.info(f"Agent received {message}")

        if isinstance(message, Assign):
            if message.provision in self.provision_passport_map:
                passport = self.provision_passport_map[message.provision]
                actor = self.managed_actors[passport.id]

                print("Agent received", message)

                # Converting assignation to Assignment
                message = Assignment(
                    assignation=message.assignation,
                    args=message.args,
                    user=message.user,
                    context=self._context,
                )

                print("Agent converted", message)
                self.managed_assignments[message.assignation] = message
                await actor.apass(message)
            else:
                logger.warning(
                    "Received assignation for a provision that is not running"
                    f"Managed: {self.provision_passport_map} Received: {message.provision}"
                )
                await self.transport.log_event(
                    AssignationEvent(
                        assignation=message.assignation,
                        kind=AssignationEventKind.CRITICAL,
                        message="Actor was no longer running or not managed",
                    )
                )

        elif isinstance(message, Interrupt):
            if message.assignation in self.managed_assignments:
                passport = self.provision_passport_map[message.provision]
                actor = self.managed_actors[passport.id]
                assignment = self.managed_assignments[message.assignation]

                # Converting unassignation to unassignment
                unass = Unassignment(assignation=message.assignation, id=assignment.id)

                await actor.apass(unass)
            else:
                logger.warning(
                    "Received unassignation for a provision that is not running"
                    f"Managed: {self.provision_passport_map} Received: {message.provision}"
                )
                await self.transport.log_event(
                    AssignationEvent(
                        assignation=message.assignation,
                        kind=AssignationEventKind.CRITICAL,
                        message="Actor could not be interupted because it was no longer running or not managed",
                    )
                )

        elif isinstance(message, AssignInquiry):
            if message.assignation in self.managed_assignments:
                passport = self.provision_passport_map[message.provision]
                actor = self.managed_actors[passport.id]
                assignment = self.managed_assignments[message.assignation]

                # Checking status
                status = await actor.is_assignment_still_running(assignment)
                if status:
                    await self.transport.log_event(
                        AssignationEvent(
                            assignation=message.assignation,
                            kind=AssignationEventKind.PROGRESS,
                            message="Actor is still running",
                        )
                    )
                else:
                    await self.transport.log_event(
                        AssignationEvent(
                            assignation=message.assignation,
                            kind=AssignationEventKind.CRITICAL,
                            message="After disconnect actor was no longer running (app was however still running)",
                        )
                    )
            else:
                await self.transport.log_event(
                    AssignationEvent(
                        assignation=message.assignation,
                        kind=AssignationEventKind.CRITICAL,
                        message="After disconnect actor was no longer managed (probably the app was restarted)",
                    )
                )

        elif isinstance(message, Cancel):
            if message.assignation in self.managed_assignments:
                passport = self.provision_passport_map[message.provision]
                actor = self.managed_actors[passport.id]
                assignment = self.managed_assignments[message.assignation]

                # Converting unassignation to unassignment
                unass = Unassignment(
                    assignation=message.assignation,
                    id=assignment.id,
                    context=self._context,
                )

                await actor.apass(unass)
            else:
                logger.warning(
                    "Received unassignation for a provision that is not running"
                    f"Managed: {self.provision_passport_map} Received: {message.provision}"
                )
                await self.transport.log_event(
                    AssignationEvent(
                        assignation=message.assignation,
                        kind=AssignationEventKind.CRITICAL,
                        message="Actor could not be canceled because it was no longer running or not managed",
                    )
                )

        elif isinstance(message, Provide):
            # TODO: Check if the provision is already running
            try:
                status = await self.acheck_status_for_provision(message)
                await self.transport.log_event(
                    ProvisionEvent(
                        provision=message.provision,
                        kind=ProvisionEventKind.ACTIVE,
                        message=f"Actor was already running {message}",
                    )
                )
            except KeyError as e:
                try:
                    await self.aspawn_actor_from_provision(message)
                except ProvisionException as e:
                    logger.error(
                        f"Error when spawing Actor for {message}", exc_info=True
                    )
                    await self.transport.log_event(
                        ProvisionEvent(
                            provision=message.provision,
                            kind=ProvisionEventKind.CRITICAL,
                            message=f"Error when spawing Actor for {message}",
                        )
                    )

        elif isinstance(message, Unprovide):
            if message.provision in self.provision_passport_map:
                passport = self.provision_passport_map[message.provision]
                actor = self.managed_actors[passport.id]
                await actor.acancel()
                await self.transport.log_event(
                    ProvisionEvent(
                        assignation=message.provision,
                        kind=ProvisionEventKind.UNHAPPY,
                        message=f"Actor was sucessfully unprovided",
                    )
                )
                del self.provision_passport_map[message.provision]
                del self.managed_actors[passport.id]
                logger.info("Actor stopped")

            else:
                await self.transport.log_event(
                    ProvisionEvent(
                        assignation=message.provision,
                        kind=ProvisionEventKind.CRITICAL,
                        message=f"Received Unprovision for never provisioned provision",
                    )
                )

        else:
            raise AgentException(f"Unknown message type {type(message)}")

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        cancelations = [actor.acancel() for actor in self.managed_actors.values()]
        # just stopping the actor, not cancelling the provision..

        for c in cancelations:
            try:
                await c
            except asyncio.CancelledError:
                pass

        await self.transport.__aexit__(exc_type, exc_val, exc_tb)

    async def aregister_definitions(self, instance_id: Optional[str] = None):
        """Registers the definitions that are defined in the definition registry

        This method is called by the agent when it starts and it is responsible for
        registering the definitions that are defined in the definition registry. This
        is done by sending the definitions to arkitekt and then storing the templates
        that are returned by arkitekt in the agent's internal data structures.

        You can implement this method in your agent subclass if you want define preregistration
        logic (like registering definitions in the definition registry).
        """

        x = await aensure_agent(
            instance_id=self.instance_id,
            extensions=[extension for extension in self.extensions.keys()],
        )

        for extension_name, extension in self.extensions.items():
            definition_registry = await extension.aretrieve_registry()
            run_cleanup = await extension.should_cleanup_on_init()

            to_be_created_templates = tuple(definition_registry.templates.values())

            created_templates = await aset_extension_templates(
                SetExtensionTemplatesInput(
                    templates=to_be_created_templates,
                    runCleanup=run_cleanup,
                    instanceId=self.instance_id,
                    extension=extension_name,
                )
            )

            for template in created_templates:
                self.interface_template_map[template.interface] = template
                self.template_interface_map[template.id] = template

    async def acheck_status_for_provision(self, provide: Provide) -> ProvisionEventKind:
        passport = self.provision_passport_map[provide.provision]
        actor = self.managed_actors[passport.id]
        return await actor.aget_status()

    async def abuild_actor_for_template(
        self, template: TemplateFragment, passport: Passport, transport: ActorTransport
    ) -> Actor:
        if not template.extension:
            raise ProvisionException(
                "No extension specified. This should not happen with the current implementation"
            )

        try:
            extension = self.extensions[template.extension]

            actor = await extension.aspawn_actor_from_template(
                template=template,
                passport=passport,
                transport=transport,
                agent=self,
                collector=self.collector,
            )
        except Exception as e:
            print(self.extensions)
            raise ProvisionException("Error spawning actor from extension") from e

        if not actor:
            raise ProvisionException("No extensions managed to spawn an actor")

        return actor

    async def on_assign_change(self, assignment: Assignment, *args, **kwargs):
        await self.transport.change_assignation(assignment.assignation, *args, **kwargs)

    async def on_assign_log(self, assignment: Assignment, *args, **kwargs):
        await self.transport.log_to_assignation(assignment.assignation, *args, **kwargs)

    async def on_actor_change(self, passport: Passport, *args, **kwargs):
        print("Changing actor state?")
        await self.transport.change_provision(passport.provision, *args, **kwargs)

    async def on_actor_log(self, passport: Passport, *args, **kwargs):
        await self.transport.log_to_provision(passport.provision, *args, **kwargs)

    async def aspawn_actor_from_provision(self, provide_message: Provide) -> Actor:
        """Spawns an Actor from a Provision. This function closely mimics the
        spawining protocol within an actor. But maps template"""

        provision = await aget_provision(
            provide_message.provision,
            rath=self.rath,
        )

        passport = Passport(provision=provision.id, instance_id=self.instance_id)

        transport = ProxyActorTransport(
            passport=passport,
            on_log_event=self.transport.log_event,
        )

        actor = await self.abuild_actor_for_template(
            provision.template, passport, transport
        )

        await actor.arun()  # TODO: Maybe move this outside?
        self.managed_actors[actor.passport.id] = actor
        self.provision_passport_map[provision.id] = actor.passport

        return actor

    async def await_errorfuture(self):
        return await self._errorfuture

    async def astep(self):
        queue_task = asyncio.create_task(self._inqueue.get(), name="queue_future")
        error_task = asyncio.create_task(self.await_errorfuture(), name="error_future")
        done, pending = await asyncio.wait(
            [queue_task, error_task],
            return_when=asyncio.FIRST_COMPLETED,
        )

        if self._errorfuture.done():
            raise self._errorfuture.exception()
        else:
            await self.process(await done.pop())

    async def astart(self, instance_id: Optional[str] = None):
        self._context = await self.hook_registry.arun_startup(self.instance_id)
        await self.hook_registry.arun_background(self._context)

        await self.aregister_definitions(instance_id=instance_id)
        self._errorfuture = asyncio.Future()
        await self.transport.aconnect(instance_id or self.instance_id)
        self.started = True

    def step(self, *args, **kwargs):
        return unkoil(self.astep, *args, **kwargs)

    def start(self, *args, **kwargs):
        return unkoil(self.astart, *args, **kwargs)

    def provide(self, *args, **kwargs):
        return unkoil(self.aprovide, *args, **kwargs)

    async def aloop(self):
        try:
            while True:
                self.running = True
                await self.astep()
        except asyncio.CancelledError:
            logger.info(
                "Provisioning task cancelled. We are running" f" {self.transport}"
            )
            self.running = False
            raise

    async def aprovide(self, instance_id: Optional[str] = None):
        logger.info(
            f"Launching provisioning task. We are running {self.transport.instance_id}"
        )
        await self.astart(instance_id=instance_id)
        logger.info("Starting to listen for requests")
        await self.aloop()

    async def __aenter__(self):
        self._inqueue = asyncio.Queue()
        self.transport.set_callback(self)
        await self.transport.__aenter__()
        return self

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from rekuest_next.errors import RekuestError


class SerializationError(RekuestError):
    """Base class for all serialization errors."""

    pass


class ExpandingError(SerializationError):
    """Base class for all expanding errors."""

    pass


class ShrinkingError(SerializationError):
    """Base class for all shrinking errors."""

    pass


class PortShrinkingError(ShrinkingError):
    """Base class for all port shrinking errors."""

    pass


class PortExpandingError(ExpandingError):
    """Base class for all port expanding errors."""

    pass


class StructureShrinkingError(PortShrinkingError):
    """Raised when a structure cannot be shrunk"""

    pass


class StructureExpandingError(PortExpandingError):
    """Raised when a structure cannot be expanded"""

    pass


class StructureRegistryError(Exception):
    """Base class for all structure registry errors."""

    pass


class StructureOverwriteError(StructureRegistryError):
    """Raised when a structure is attempted to be overwritten in the registry"""

    pass


class StructureDefinitionError(StructureRegistryError):
    """Raised when a structure was never defined in the registry"""

    pass



from rekuest_next.errors import RekuestError


class DefinitionError(RekuestError):
    """Base class for all definition errors"""

    pass


class NoDefinitionRegistryFound(RekuestError):
    """Raised when no definition registry is found"""

    pass


class NonSufficientDocumentation(DefinitionError):
    """Raised when we cannot infer sufficcient documentatoin"""



from rekuest_next.actors.vars import (
    get_current_assignation_helper,
)
from rekuest_next.api.schema import LogLevel
from typing import Dict, Any, Optional


async def alog(message: str, level: LogLevel = LogLevel.DEBUG) -> None:
    await get_current_assignation_helper().alog(level, message)


def log(message: str, level: LogLevel = LogLevel.DEBUG) -> None:
    get_current_assignation_helper().log(level, message)


def useUser() -> str:
    """Returns the user id of the current assignation"""
    return get_current_assignation_helper().user


def useInstanceID() -> str:
    """Returns the guardian id of the current provision"""
    return get_current_assignation_helper().passport.instance_id


def useContext(value: Optional[str] = None) -> Dict[str, Any]:
    context = get_current_assignation_helper().assignment.context
    if value:
        context = context[value]

    return context


def progress(percentage: int) -> None:
    """Progress

    Args:
        percentage (int): Percentage to progress to
    """

    helper = get_current_assignation_helper()
    helper.progress(percentage)


async def aprogress(percentage: int) -> None:
    """Progress

    Args:
        percentage (int): Percentage to progress to
    """

    helper = get_current_assignation_helper()
    await helper.aprogress(percentage)



from rekuest_next.errors import RekuestError


class PostmanException(RekuestError):
    """
    Base class for all exceptions raised by the Agent.
    """


class AssignException(PostmanException):
    """
    Raised when an error occurs during the assignment of a task to an agent.
    """


class RecoverableAssignException(AssignException):
    """
    Raised when an error occurs during the assignment of a task to an agent.
    """


class IncorrectReserveState(AssignException):
    """
    Raised when a assignation during an incorect reservation state for this contract
    """



from rekuest_next.structures.registry import StructureRegistry
from rekuest_next.structures.hooks.standard import id_shrink
from rekuest_next.api.schema import (
    TemplateFragment,
    NodeFragment,
    Search_templatesQuery,
    Search_nodesQuery,
    Search_testcasesQuery,
    Search_testresultsQuery,
    TestCaseFragment,
    TestResultFragment,
    aget_template,
    aget_testcase,
    aget_testresult,
    aget_template,
    afind,
    PortScope,
)
from rekuest_next.widgets import SearchWidget


DEFAULT_STRUCTURE_REGISTRY = None


def get_default_structure_registry() -> StructureRegistry:
    global DEFAULT_STRUCTURE_REGISTRY
    if not DEFAULT_STRUCTURE_REGISTRY:
        DEFAULT_STRUCTURE_REGISTRY = StructureRegistry()

        DEFAULT_STRUCTURE_REGISTRY.register_as_structure(
            TemplateFragment,
            "@rekuest_next/template",
            scope=PortScope.GLOBAL,
            aexpand=aget_template,
            ashrink=id_shrink,
            default_widget=SearchWidget(
                query=Search_templatesQuery.Meta.document, ward="rekuest_next"
            ),
        )

        DEFAULT_STRUCTURE_REGISTRY.register_as_structure(
            NodeFragment,
            "@rekuest_next/node",
            scope=PortScope.GLOBAL,
            aexpand=afind,
            ashrink=id_shrink,
            default_widget=SearchWidget(
                query=Search_nodesQuery.Meta.document, ward="rekuest_next"
            ),
        )

        DEFAULT_STRUCTURE_REGISTRY.register_as_structure(
            TestCaseFragment,
            "@rekuest_next/testcase",
            scope=PortScope.GLOBAL,
            aexpand=aget_testcase,
            ashrink=id_shrink,
            default_widget=SearchWidget(
                query=Search_testcasesQuery.Meta.document, ward="rekuest_next"
            ),
        )

        DEFAULT_STRUCTURE_REGISTRY.register_as_structure(
            TestResultFragment,
            "@rekuest_next/testresult",
            scope=PortScope.GLOBAL,
            aexpand=aget_testresult,
            ashrink=id_shrink,
            default_widget=SearchWidget(
                query=Search_testresultsQuery.Meta.document, ward="rekuest_next"
            ),
        )

    return DEFAULT_STRUCTURE_REGISTRY



__version__ = "0.1.1"




from .utils import acall, afind, areserve, reserved, call, aiterate, iterate

from .arkitekt import imported


__all__ = [
    "acall",
    "afind",
    "areserve",
    "reserved",
    "call",
    "find",
    "reserve",
    "structur_reg",
    "imported",
    "iterate",
    "aiterate",
]



from rekuest_next.agents.errors import AgentException


class AgentTransportException(AgentException):
    """
    Base class for all exceptions raised by the Agent Transport.
    """


class ProvisionListDeniedError(AgentTransportException):
    """
    Raised when the backend is not able to list the provisions.
    """


class AssignationListDeniedError(AgentTransportException):
    """
    Raised when the backend is not able to list the assignations.
    """


class CorrectableConnectionFail(AgentTransportException):
    pass


class DefiniteConnectionFail(AgentTransportException):
    pass


class AgentConnectionFail(AgentTransportException):
    pass


class AgentWasKicked(AgentConnectionFail):
    pass


class AgentWasBlocked(AgentConnectionFail):
    pass


class AgentIsAlreadyBusy(AgentConnectionFail):
    pass



from typing import Protocol, Optional, List
from rekuest_next.api.schema import (
    AssignWidgetInput,
    ReturnWidgetInput,
    PortInput,
    PortScope,
)
from pydantic import BaseModel
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Optional,
    Type,
    TypeVar,
    Protocol,
    runtime_checkable,
)


class PortBuilder(Protocol):
    def __call__(
        self,
        cls: type,
        assign_widget: Optional[AssignWidgetInput],
        return_widget: Optional[ReturnWidgetInput],
    ) -> PortInput: ...


class FullFilledStructure(BaseModel):
    fullfilled_by: str
    cls: Type
    identifier: str
    scope: PortScope
    aexpand: Callable[
        [
            str,
        ],
        Awaitable[Any],
    ]
    ashrink: Callable[
        [
            any,
        ],
        Awaitable[str],
    ]
    acollect: Callable[
        [
            str,
        ],
        Awaitable[Any],
    ]
    predicate: Callable[[Any], bool]
    convert_default: Callable[[Any], str]
    default_widget: Optional[AssignWidgetInput]
    default_returnwidget: Optional[ReturnWidgetInput]

    class Config:
        arbitrary_types_allowed = True
        copy_on_model_validation = False
        extra = "forbid"


class FullFilledArg(BaseModel):
    key: str
    default: Optional[Any]
    cls: Any
    description: Optional[str]


class FullFilledModel(BaseModel):
    identifier: str
    description: Optional[str]
    args: List[FullFilledArg]



from typing import Any, List, Optional
import uuid
from typing_extensions import Literal


from pydantic import BaseModel, Field
from datetime import datetime
from rekuest_nextmessages import Assignation, Reservation, Unassignation, Unreservation
from rekuest_nextapi.schema import ReservationStatus, AssignationStatus, ReserveParamsInput
from enum import Enum


class PostmanMessageTypes(str, Enum):
    LIST_RESERVATION = "RESERVE_LIST"
    LIST_RESERVATION_REPLY = "RESERVE_LIST_REPLY"
    LIST_RESERVATION_DENIED = "RESERVE_LIST_DENIED"

    RESERVE = "RESERVE"
    RESERVE_REPLY = "RESERVE_REPLY"
    RESERVE_DENIED = "RESERVE_DENIED"

    UNRESERVE = "UNRESERVE"
    UNRESERVE_REPLY = "UNRESERVE_REPLY"
    UNRESERVE_DENIED = "UNRESERVE_DENIED"

    LIST_ASSIGNATION = "ASSIGN_LIST"
    LIST_ASSIGNATION_REPLY = "ASSIGN_LIST_REPLY"
    LIST_ASSIGNATION_DENIED = "ASSIGN_LIST_DENIED"

    ASSIGN = "ASSIGN"
    ASSIGN_REPLY = "ASSIGN_REPLY"
    ASSIGN_DENIED = "ASSIGN_DENIED"

    UNASSIGN = "UNASSIGN"
    UNASSIGN_REPLY = "UNASSIGN_REPLY"
    UNASSIGN_DENIED = "UNASSIGN_DENIED"


class PostmanSubMessageTypes(str, Enum):
    ASSIGN_UPDATE = "ASSIGN_UPDATE"
    RESERVE_UPDATE = "RESERVE_UPDATE"


class JSONMeta(BaseModel):
    timestamp: datetime = Field(default_factory=datetime.utcnow)


class JSONMessage(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    type: str
    meta: JSONMeta = Field(default_factory=JSONMeta)


class ReserveList(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_RESERVATION
    ] = PostmanMessageTypes.LIST_RESERVATION
    exclude: Optional[List[ReservationStatus]]


class ReserveListReply(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_RESERVATION_REPLY
    ] = PostmanMessageTypes.LIST_RESERVATION_REPLY
    reservations: List[Reservation]


class ReserveListDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_RESERVATION_DENIED
    ] = PostmanMessageTypes.LIST_RESERVATION_DENIED
    error: str


class ReservePub(JSONMessage):
    type: Literal[PostmanMessageTypes.RESERVE] = PostmanMessageTypes.RESERVE
    provision: Optional[str]
    params: Optional[ReserveParamsInput]
    node: Optional[str]
    template: Optional[str]
    title: Optional[str]
    reference: Optional[str]


class ReservePubReply(JSONMessage, Reservation):
    type: Literal[PostmanMessageTypes.RESERVE_REPLY] = PostmanMessageTypes.RESERVE_REPLY


class ReservePubDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.RESERVE_DENIED
    ] = PostmanMessageTypes.RESERVE_DENIED
    error: str


class UnreservePub(JSONMessage):
    type: Literal[PostmanMessageTypes.UNRESERVE] = PostmanMessageTypes.UNRESERVE
    reservation: str


class UnreservePubReply(JSONMessage, Unreservation):
    type: Literal[
        PostmanMessageTypes.UNRESERVE_REPLY
    ] = PostmanMessageTypes.UNRESERVE_REPLY


class UnreservePubDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.UNRESERVE_DENIED
    ] = PostmanMessageTypes.UNRESERVE_DENIED
    error: str


class ReserveSubUpdate(JSONMessage, Reservation):
    type: Literal[
        PostmanSubMessageTypes.RESERVE_UPDATE
    ] = PostmanSubMessageTypes.RESERVE_UPDATE


class AssignList(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_ASSIGNATION
    ] = PostmanMessageTypes.LIST_ASSIGNATION
    exclude: Optional[List[AssignationStatus]]


class AssingListReply(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_ASSIGNATION_REPLY
    ] = PostmanMessageTypes.LIST_ASSIGNATION_REPLY
    assignations: List[Assignation]


class AssignListDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.LIST_ASSIGNATION_DENIED
    ] = PostmanMessageTypes.LIST_ASSIGNATION_DENIED
    error: str


class AssignPub(JSONMessage):
    type: Literal[PostmanMessageTypes.ASSIGN] = PostmanMessageTypes.ASSIGN
    reservation: str
    reference: Optional[str]
    parent: Optional[str]
    args: List[Any]
    persist: bool = True
    log: bool = True


class AssignPubReply(JSONMessage, Assignation):
    type: Literal[PostmanMessageTypes.ASSIGN_REPLY] = PostmanMessageTypes.ASSIGN_REPLY
    assignation: str
    status: AssignationStatus


class AssignPubDenied(JSONMessage):
    type: Literal[PostmanMessageTypes.ASSIGN_DENIED] = PostmanMessageTypes.ASSIGN_DENIED
    error: str


class UnassignPub(JSONMessage):
    type: Literal[PostmanMessageTypes.UNASSIGN] = PostmanMessageTypes.UNASSIGN
    assignation: str


class UnassignPubReply(JSONMessage, Unassignation):
    type: Literal[
        PostmanMessageTypes.UNASSIGN_REPLY
    ] = PostmanMessageTypes.UNASSIGN_REPLY


class UnassignPubDenied(JSONMessage):
    type: Literal[
        PostmanMessageTypes.UNASSIGN_DENIED
    ] = PostmanMessageTypes.UNASSIGN_DENIED
    error: str


class AssignSubUpdate(JSONMessage, Assignation):
    type: Literal[
        PostmanSubMessageTypes.ASSIGN_UPDATE
    ] = PostmanSubMessageTypes.ASSIGN_UPDATE



from random import random
from typing import (
    Awaitable,
    Callable,
    Coroutine,
    Optional,
    Union,
    TypeVar,
    runtime_checkable,
    Protocol,
    Any,
    Dict,
    List,
    AsyncIterator,
)
import uuid
from pydantic import Field
from fluss_next.api.schema import BindsInput
from rekuest_next.messages import AssignationEvent, ProvisionEvent
from rekuest_next.structures.default import get_default_structure_registry
from koil.composition import KoiledModel
from koil.types import ContextBool
import uuid
import asyncio
import logging
from rekuest_next.structures.registry import StructureRegistry
from rekuest_next.api.schema import (
    DefinitionFragment,
    ReservationFragment,
    TemplateFragment,
    afind,
)
from .errors import (
    AssignException,
    IncorrectReserveState,
    PostmanException,
    RecoverableAssignException,
)
from rekuest_next.actors.base import Actor, SerializingActor
from rekuest_next.agents.transport.base import AgentTransport
from .base import BasePostman
from rekuest_next.actors.types import (
    Passport,
    Assignment,
    Unassignment,
    AssignmentUpdate,
)
from enum import Enum

logger = logging.getLogger(__name__)

T = TypeVar("T")


class ContractStatus(str, Enum):
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"


@runtime_checkable
class ContractStateHook(Protocol):
    async def __call__(
        self, state: ContractStatus = None, reference: str = None
    ) -> None: ...


@runtime_checkable
class RPCContract(Protocol):
    async def __aenter__(self: Any) -> Any: ...

    async def __aexit__(self, exc_type, exc, tb): ...

    async def change_state(self, state: ContractStatus): ...

    async def aassign(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        reference: Optional[str] = None,
        assign_timeout: Optional[float] = None,
    ) -> Dict[str, Any]: ...

    async def aassign_retry(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        reference: Optional[str] = None,
        assign_timeout: Optional[float] = None,
        retry: Optional[int] = 0,
    ) -> Dict[str, Any]: ...

    async def astream(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        reference: Optional[str] = None,
        yield_timeout: Optional[float] = None,
    ) -> AsyncIterator[List[Any]]: ...

    async def astream_retry(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        reference: Optional[str] = None,
        yield_timeout: Optional[float] = None,
        retry: Optional[int] = 0,
    ) -> Dict[str, Any]: ...


class RPCContractBase(KoiledModel):
    max_retries: int = 3
    retry_delay_ms: float = 1000
    reference: Optional[str]
    active: ContextBool = Field(default=False)
    state: ContractStatus = Field(default=ContractStatus.INACTIVE)
    state_hook: Optional[ContractStateHook] = None
    timeout_is_recoverable: bool = True

    async def aenter(self):
        raise NotImplementedError("Should be implemented by subclass")

    async def aexit(self):
        raise NotImplementedError("Should be implemented by subclass")

    async def change_state(self, state: ContractStatus):
        self.state = state
        if self.state_hook:
            await self.state_hook(state=state, reference=self.reference)

    async def aassign(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        reference: Optional[str] = None,
        assign_timeout: Optional[int] = None,
    ):
        raise NotImplementedError("Should be implemented by subclass")

    async def astream(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        reference: Optional[str] = None,
        yield_timeout: Optional[int] = None,
    ):
        raise NotImplementedError("Should be implemented by subclass")

    async def astream_retry(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        yield_timeout: Optional[int] = None,
        retry: Optional[int] = 0,
        reference: Optional[str] = None,
        retry_delay_ms: Optional[int] = None,
    ) -> Dict[str, Any]:
        try:
            async for i in self.astream(
                kwargs={**kwargs},
                parent=parent,
                reference=reference,
                yield_timeout=yield_timeout,
            ):
                yield i
        except RecoverableAssignException as e:
            logger.warning(f"Stream failed with {e}")
            if retry < self.max_retries:
                await asyncio.sleep((retry_delay_ms or self.retry_delay_ms) * 0.001)
                async for i in self.astream_retry(
                    kwargs=kwargs,
                    parent=parent,
                    reference=reference,
                    yield_timeout=yield_timeout,
                    retry=retry + 1,
                    retry_delay_ms=retry_delay_ms,
                ):
                    yield i
            else:
                raise e

    async def aassign_retry(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        assign_timeout: Optional[int] = None,
        retry: Optional[int] = 0,
        reference: Optional[str] = None,
        retry_delay_ms: Optional[int] = None,
    ) -> Dict[str, Any]:
        try:
            return await self.aassign(
                kwargs={**kwargs},
                parent=parent,
                reference=reference,
                assign_timeout=assign_timeout,
            )
        except RecoverableAssignException as e:
            logger.warning(f"Assign failed with {e}")
            if retry < self.max_retries:
                logger.info(f"Retrying in {retry_delay_ms or self.retry_delay_ms}ms")
                await asyncio.sleep((retry_delay_ms or self.retry_delay_ms) * 0.001)
                return await self.aassign_retry(
                    kwargs=kwargs,
                    parent=parent,
                    reference=reference,
                    assign_timeout=assign_timeout,
                    retry=retry + 1,
                    retry_delay_ms=retry_delay_ms,
                )
            else:
                raise e

    async def __aenter__(self: T) -> T:
        await self.aenter()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.aexit()


class actoruse(RPCContractBase):
    template: TemplateFragment
    supervisor: Actor = Field(repr=False, exclude=True)
    reference: Optional[str]
    "The governing actor"
    assign_timeout: Optional[float] = 36000
    yield_timeout: Optional[float] = 2000

    _transport: AgentTransport = None
    _actor: SerializingActor
    _enter_future: asyncio.Future = None
    _exit_future: asyncio.Future = None
    _updates_queue: asyncio.Queue[Union[AssignationEvent, ProvisionEvent]] = None
    _updates_watcher: asyncio.Task = None
    _assign_queues = {}

    async def aassign(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        assign_timeout: Optional[float] = None,
        reference: Optional[str] = None,
    ) -> Dict[str, Any]:
        assignment = Assignment(
            assignation=parent.assignation if parent else None,
            parent=parent.id if parent else None,
            args=serialize_inputs(self._actor.definition, kwargs),
            status=AssignationStatus.ASSIGNED,
            user=parent.user if parent else None,
            reference=reference,
        )

        _ass_queue = asyncio.Queue[AssignmentUpdate]()
        self._assign_queues[assignment.id] = _ass_queue

        await self._actor.apass(assignment)
        try:
            while True:  # Waiting for assignation
                logger.info("Waiting for update")
                ass = await asyncio.wait_for(
                    _ass_queue.get(), timeout=assign_timeout or self.assign_timeout
                )
                logger.info(f"Local Assign Context: {ass}")
                if ass.status == AssignationStatus.RETURNED:
                    return deserialize_outputs(self._actor.definition, ass.returns)

                if ass.status in [AssignationStatus.ERROR]:
                    raise RecoverableAssignException(
                        f"Recoverable Exception: {ass.message}"
                    )

                if ass.status in [AssignationStatus.CRITICAL]:
                    raise AssignException(f"Critical error: {ass.message}")

                if ass.status in [AssignationStatus.CANCELLED]:
                    raise AssignException("Was cancelled from the outside")

                _ass_queue.task_done()
        except asyncio.CancelledError as e:
            await self._actor.apass(
                Unassignation(
                    assignation=id,
                )
            )

            ass = await asyncio.wait_for(_ass_queue.get(), timeout=2)
            if ass.status == AssignationStatus.CANCELING:
                logger.info("Wonderfully cancelled that assignation!")
                raise e

            raise AssignException(f"Critical error: {ass}")

        except asyncio.TimeoutError as e:
            exc_class = (
                RecoverableAssignException
                if self.timeout_is_recoverable
                else AssignException
            )

            raise exc_class("Timeout error for assignation") from e

        except Exception as e:
            logger.error("Error in Assignation", exc_info=True)
            raise e

    async def on_actor_log(self, *args, **kwargs):
        logger.info(f"ActorLog: {args} {kwargs}")

    async def on_assign_log(self, *args, **kwargs):
        logger.info(f"AssingLog: {args} {kwargs}")

    async def on_actor_change(self, passport: Passport, status, **kwargs):
        # passport is irellevant because we only every manage one actor
        if status == ProvisionStatus.ACTIVE:
            await self.change_state(ContractStatus.ACTIVE)
            if self._enter_future and not self._enter_future.done():
                self._enter_future.set_result(True)

        else:
            await self.change_state(ContractStatus.INACTIVE)
            if self._enter_future and not self._enter_future.done():
                self._enter_future.set_exception(Exception("Error on provision"))

    async def astream(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        yield_timeout: Optional[float] = None,
        reference: Optional[str] = None,
    ) -> AsyncIterator[Dict[str, Any]]:
        inputs = serialize_inputs(self._actor.definition, kwargs)
        assignment = Assignment(
            assignation=parent.assignation if parent else None,
            parent=parent.id if parent else None,
            args=inputs,
            status=AssignationStatus.ASSIGNED,
            user=parent.user if parent else None,
            reference=reference,
        )

        _ass_queue = asyncio.Queue[AssignmentUpdate]()
        self._assign_queues[assignment.id] = _ass_queue

        await self._actor.apass(assignment)

        try:
            while True:  # Waiting for assignation
                ass = await asyncio.wait_for(
                    _ass_queue.get(), timeout=yield_timeout or self.yield_timeout
                )
                logger.info(f"Local Stream Context: {ass}")
                if ass.status == AssignationStatus.YIELD:
                    yield deserialize_outputs(self._actor.definition, ass.returns)

                if ass.status == AssignationStatus.DONE:
                    return

                if ass.status in [AssignationStatus.CRITICAL, AssignationStatus.ERROR]:
                    raise AssignException(f"Critical error: {ass.message}")

                _ass_queue.task_done()

        except asyncio.CancelledError as e:
            await self._actor.apass(
                Unassignment(assignation=assignment.id, id=assignment.id)
            )

            ass = await asyncio.wait_for(_ass_queue.get(), timeout=2)
            if ass.status == AssignationStatus.CANCELING:
                logger.info("Wonderfully cancelled that assignation!")
                raise e

            raise e

        except asyncio.TimeoutError as e:
            exc_class = (
                RecoverableAssignException
                if self.timeout_is_recoverable
                else AssignException
            )

            raise exc_class("Timeout error for assignation") from e

        except Exception as e:
            logger.error(exc_info=True)
            raise e

    async def on_assign_change(
        self,
        assignment: Assignment,
        status=None,
        returns=None,
        progress=None,
        message=None,
    ):
        await self._assign_queues[assignment.id].put(
            AssignmentUpdate(
                assignment=assignment.id,
                status=status,
                returns=returns,
                progress=progress,
                message=message,
            )
        )

        return

    async def aenter(self):
        self._enter_future = asyncio.Future()
        self._updates_queue = asyncio.Queue[AssignationChangedMessage]()

        self._actor = await self.supervisor.aspawn_actor(
            self.template,
            on_actor_log=self.on_actor_log,
            on_actor_change=self.on_actor_change,
            on_assign_change=self.on_assign_change,
            on_assign_log=self.on_assign_log,
        )

        await self._actor.arun()
        await self._enter_future

    async def aexit(self):
        if self._actor:
            await self._actor.acancel()

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        copy_on_model_validation = False


class arkiuse(RPCContractBase):
    hash: Optional[str] = None
    provision: Optional[str] = None
    reference: str = "default"
    binds: Optional[BindsInput] = None
    postman: BasePostman = Field(repr=False)
    reserve_timeout: Optional[int] = 100000
    assign_timeout: Optional[int] = 100000
    yield_timeout: Optional[int] = 100000
    auto_unreserve: bool = False

    _reservation: ReservationFragment = None
    _enter_future: asyncio.Future = None
    _exit_future: asyncio.Future = None
    _updates_queue: asyncio.Queue = None
    _updates_watcher: asyncio.Task = None
    _definition: Optional[DefinitionFragment] = None

    async def aassign(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        assign_timeout: Optional[int] = None,
        reference: Optional[str] = None,
    ) -> Dict[str, Any]:
        assert self._reservation, "We never entered the context manager"
        if self.state != ContractStatus.ACTIVE:
            raise IncorrectReserveState(
                f"Contract is not active at the moment: {self.state}"
            )

        inputs = serialize_inputs(self._definition, kwargs)

        try:
            _ass_queue = await self.postman.aassign(
                self._reservation.id,
                inputs,
                parent=parent.assignation if parent else None,
                reference=reference,
            )
        except PostmanException as e:
            raise AssignException("Cannot do initial assignment") from e

        ass = None
        try:
            while True:  # Waiting for assignation
                ass = await asyncio.wait_for(
                    _ass_queue.get(), timeout=assign_timeout or self.assign_timeout
                )
                logger.info(f"Assign Context: {ass}")
                if ass.status == AssignationStatus.RETURNED:
                    return deserialize_outputs(self._definition, ass.returns)

                if ass.status in [AssignationStatus.ERROR]:
                    raise RecoverableAssignException(
                        f"Recoverable Exception: {ass.statusmessage}"
                    )

                if ass.status in [AssignationStatus.CRITICAL]:
                    raise AssignException(f"Critical error: {ass.statusmessage}")

                if ass.status in [AssignationStatus.CANCELLED]:
                    raise AssignException("Was cancelled from the outside")

        except asyncio.CancelledError as e:
            if ass:
                await self.postman.aunassign(ass.id)

                ass = await asyncio.wait_for(_ass_queue.get(), timeout=2)
                if ass.status == AssignationStatus.CANCELING:
                    logger.info("Wonderfully cancelled that assignation!")
                    raise e

                raise PostmanException(
                    f"Unexpected Arkitekt repsonse while trying to cancel exception: {ass}"
                )

        except asyncio.TimeoutError as e:
            if ass:
                logger.warning(
                    f"Cancelling this assignation but not wait for request {ass}"
                )
                await self.postman.aunassign(ass.id)

            exc_class = (
                RecoverableAssignException
                if self.timeout_is_recoverable
                else AssignException
            )

            raise exc_class("Timeout error for assignation") from e

    async def astream(
        self,
        kwargs: Dict[str, Any],
        parent: Optional[Assignment] = None,
        yield_timeout: Optional[int] = None,
        reference: Optional[str] = None,
    ) -> AsyncIterator[Dict[str, Any]]:
        assert self._reservation, "We never entered the context manager"
        if self.state != ContractStatus.ACTIVE:
            raise IncorrectReserveState(
                f"Contract is not active at the moment: {self.state}"
            )

        try:
            _ass_queue = await self.postman.aassign(
                self._reservation.id,
                serialize_inputs(self._definition, kwargs),
                parent=parent.assignation if parent else None,
                reference=reference,
            )
        except PostmanException as e:
            raise AssignException("Cannot do initial assignment") from e
        ass = None

        try:
            while True:  # Waiting for assignation
                ass = await asyncio.wait_for(
                    _ass_queue.get(), timeout=yield_timeout or self.yield_timeout
                )
                logger.info(f"Stream Context: {ass}")
                if ass.status == AssignationStatus.YIELD:
                    yield deserialize_outputs(self._definition, ass.returns)

                if ass.status in [AssignationStatus.DONE]:
                    return

                if ass.status in [AssignationStatus.ERROR]:
                    raise RecoverableAssignException(
                        f"Recoverable Exception: {ass.statusmessage}"
                    )

                if ass.status in [AssignationStatus.CRITICAL]:
                    raise AssignException(f"Critical error: {ass.statusmessage}")

                if ass.status in [AssignationStatus.CANCELLED]:
                    raise AssignException("Was cancelled from the outside")

        except asyncio.CancelledError as e:
            if ass:
                logger.warning(f"Cancelling this assignation {ass}")
                await self.postman.aunassign(ass.id)

                ass = await asyncio.wait_for(_ass_queue.get(), timeout=2)
                if ass.status == AssignationStatus.CANCELING:
                    logger.info("Wonderfully cancelled that assignation!")
                    raise e

                raise PostmanException(
                    f"Unexpected Arkitekt repsonse while trying to cancel exception: {ass}"
                )

        except asyncio.TimeoutError as e:
            if ass:
                logger.warning(
                    f"Cancelling this assignation but not wait for request {ass}"
                )
                await self.postman.aunassign(ass.id)

            exc_class = (
                RecoverableAssignException
                if self.timeout_is_recoverable
                else AssignException
            )

            raise exc_class("Timeout error for assignation") from e

    async def watch_updates(self):
        logger.info("Waiting for updates")
        try:
            while True:
                self._reservation = await self._updates_queue.get()
                logger.info(f"Updated Reservation {self._reservation}")
                if self._reservation.status == ReservationStatus.ROUTING:
                    logger.info("Reservation is routing")

                elif self._reservation.status == ReservationStatus.ACTIVE:
                    if self._enter_future and not self._enter_future.done():
                        logger.info("Entering future")
                        self._enter_future.set_result(True)

                    await self.change_state(ContractStatus.ACTIVE)

                elif self._reservation.status == ReservationStatus.DISCONNECT:
                    if self._enter_future and not self._enter_future.done():
                        logger.info("Entering future")
                        self._enter_future.set_result(True)

                    await self.change_state(ContractStatus.INACTIVE)

                else:
                    logger.error(
                        f"Currently unhandled status {self._reservation.status}"
                    )
                    if self._enter_future and not self._enter_future.done():
                        self._enter_future.set_exception(True)

                    await self.change_state(ContractStatus.INACTIVE)

        except asyncio.CancelledError:
            pass

    async def aenter(self):
        logger.info(f"Trying to reserve {self.hash}")

        self._enter_future = asyncio.Future()
        self._definition = await afind(hash=self.hash)
        self._updates_queue = await self.postman.areserve(
            hash=self.hash,
            params=self.params,
            provision=self.provision,
            reference=self.reference,
            binds=self.binds,
        )
        try:
            self._updates_watcher = asyncio.create_task(self.watch_updates())
            await asyncio.wait_for(
                self._enter_future, self.reserve_timeout
            )  # Waiting to enter

        except asyncio.TimeoutError:
            logger.warning("Reservation timeout")
            self._updates_watcher.cancel()

            try:
                await self._updates_watcher
            except asyncio.CancelledError:
                pass

            raise

        return self

    async def aexit(self):
        self.active = False

        if self._reservation:
            if self.auto_unreserve:
                logger.info(f"Unreserving {self.hash}")
                await self.postman.aunreserve(self._reservation.id)

        if self._updates_watcher:
            self._updates_watcher.cancel()

            try:
                await self._updates_watcher
            except asyncio.CancelledError:
                pass

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        copy_on_model_validation = False


class mockuse(RPCContract):
    returns: tuple = (1,)
    streamevents: int = 3
    assign_sleep: float = Field(default_factory=random)
    reserve_sleep: float = Field(default_factory=random)
    unreserve_sleep: float = Field(default_factory=random)
    stream_sleep: float = Field(default_factory=random)

    async def aenter(self):
        await asyncio.sleep(self.reserve_sleep)
        self.active = True
        return self

    async def aexit(self):
        self.active = False
        await asyncio.sleep(self.unreserve_sleep)

    async def aassign(
        self,
        *args,
        structure_registry=None,
        alog,
        **kwargs,
    ):
        assert self.active, "We never entered the contract"
        if alog:
            await alog(
                Assignation(assignation=str(uuid.uuid4())),
                AssignationLogLevel.INFO,
                "Mock assignation",
            )
        await asyncio.sleep(self.assign_sleep)
        return self.returns

    async def astream(
        self,
        *args,
        structure_registry=None,
        alog,
        **kwargs,
    ):
        assert self.active, "We never entered the contract"
        if alog:
            await alog(
                Assignation(assignation=str(uuid.uuid4())),
                AssignationLogLevel.INFO,
                "Mock assignation",
            )
        for i in range(self.streamevents):
            await asyncio.sleep(self.stream_sleep)
            yield self.returns

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True


class serializingarkiuse(arkiuse):
    structure_registry: StructureRegistry = Field(
        default_factory=get_default_structure_registry, repr=False
    )

    async def aassign(
        self,
        *args,
        parent: Optional[Assignment] = None,
        assign_timeout: Optional[int] = None,
        reference: Optional[str] = None,
        **kwargs,
    ) -> Coroutine[Any, Any, Dict[str, Any]]:
        shrinked_kwargs = await shrink_inputs(
            self._definition, args, kwargs, self.structure_registry
        )

        unshrunk = await super().aassign(
            shrinked_kwargs, parent, assign_timeout, reference
        )

        return await expand_outputs(self._definition, unshrunk, self.structure_registry)



from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Optional,
    Type,
    TypeVar,
    Protocol,
    runtime_checkable,
    TYPE_CHECKING,
)
from pydantic import BaseModel
import inspect
from rekuest_next.structures.types import FullFilledStructure
from rekuest_next.api.schema import PortScope, AssignWidgetInput, ReturnWidgetInput


@runtime_checkable
class RegistryHook(Protocol):
    """A hook that can be registered to the structure registry
    and will be called when a structure is about to be registered
    and can be used to modify the structure with the registry

    """

    def is_applicable(
        self,
        cls: Type,
    ) -> bool:
        """Given a class, return True if this hook is applicable to it"""
        ...

    def apply(
        self,
        cls: Type,
        identifier: str = None,
        scope: PortScope = PortScope.LOCAL,
        aexpand: Callable[
            [
                str,
            ],
            Awaitable[Any],
        ] = None,
        ashrink: Callable[
            [
                any,
            ],
            Awaitable[str],
        ] = None,
        acollect: Callable[
            [
                str,
            ],
            Awaitable[Any],
        ] = None,
        predicate: Callable[[Any], bool] = None,
        convert_default: Callable[[Any], str] = None,
        default_widget: Optional[AssignWidgetInput] = None,
        default_returnwidget: Optional[ReturnWidgetInput] = None,
    ) -> FullFilledStructure:
        """App a class, return True if this hook is applicable to it"""
        ...



from typing import List, Union, Any, AsyncGenerator

from pydantic import Field

from rekuest_next.api.schema import (
    AssignationFragment,
    BindsInput,
    AssignInput,
    CancelInput,
    InterruptInput,
    ReserveInput,
    UnreserveInput,
    ReservationFragment,
    AssignationEventFragment,
)
from koil.composition import KoiledModel
import asyncio


class BasePostman(KoiledModel):
    """Postman


    Postmans are the schedulers of the arkitekt platform, they are taking care
    of the communication between your app and the arkitekt-server. And
    provide abstractions to map the asynchornous message-based nature of the arkitekt-server to
    the (a)sync nature of your app. It maps assignations to functions or generators
    depending on the definition, to mimic an actor-like behaviour.

    """

    connected = Field(default=False)
    instance_id: str

    async def aassign(
        self, input: AssignInput
    ) -> AsyncGenerator[AssignationEventFragment, None]:
        """Idea"""
        yield

    async def areserve(self, input: ReserveInput) -> ReservationFragment: ...

    async def aunreserve(self, input: UnreserveInput) -> ReservationFragment: ...
